#include <stdlib.h>                                         //instrukcje operacji na pamieci
                                                            //  malloc i free
#include <stdio.h>                                          //zmienna FILE i instrukcje operacji
                                                            //  na plikach
#include <string.h>                                         //operacje na stringach
#include "mpt_player.h"                                     //header tej biblioteki

#ifdef MPT_COMPILE                                          //czy kompilowac jednostke

#define LO_B(value) (value&0x0f)                            //mlodsza polowa bajtu
#define HI_B(value) (value>>4)                              //starsza polowa bajtu

/**********************************************************
 struktury informacji o songu
 **********************************************************/

struct MPTInstrPosT{                                        //struktura danych jednego slupku
                                                            //    instrumentu
  zbyte audc;                                               //  glosnosc glowna i brzmienie
  zbyte accent;                                             //  dane akcentu
};

struct MPTInstrT{                                           //struktura danych instrumentu
  MPTInstrPosT value[16];                                   //  16 slupkow
                                                            //  8 parametrow instrumentu
  zbyte effect;                                             //    numer efektu i jego parametr
  zbyte freqtable;                                          //    tablica czestotliwosci i
                                                            //      opoznienie dla efektu
  zbyte decvolumespeed;                                     //    predkosc opadania glosnosci
                                                            //      dzwieku
  zbyte audctl;                                             //    AUDCTL kanalu
  zbyte transpose[4];                                       //    transpozycje dzwieku
  zbyte AccParam[8];                                        //  8 parametrow akcentow
};

struct MPTTrackT{                                           //struktura danych jednego wpisu w
                                                            //    tracku
  zbyte pattern;                                            //  numer patternu do odegrania
  zbyte transpose;                                          //  transpozycja patternu
};

/**********************************************************
 zmienne informacji o songu
 **********************************************************/

FILE    *MPTFile;                                           //aktualnie odtwarzany plik MPT
zint32  MPTBegin;                                           //poczatek pliku w formacie Atari
zint32  MPTEnd;                                             //koniec pliku w formacie Atari
zint32  MPTSize;                                            //wielkosc pliku w formacie Atari
zvoid   *MPTData=NULL;                                      //adres do poczatku danych pliku
MPTInstrT *(MPTInstrument[0x20]);                           //tablica wskaznikow do danych
                                                            //  instrumentow
zbyte *(MPTPattern[0x40]);                                  //tablica wskaznikow do danych
                                                            //  patternow
zbyte *(MPTFreqTable[4]);                                   //Tablice zamian nut na czestotliwosci
MPTTrackT *(MPTTrack[4]);                                   //tablica wskaznikow do danych
                                                            //  trackow
zbyte MPTPatternLength;                                     //dlugosc patternow
zbyte MPTTempo;                                             //tempo utworu
zchar MPTSongName[MAX_PATH];                                //nazwa utworu

/**********************************************************
 zmienne pomocnicze
 **********************************************************/

zbyte MPTEmptyInstrument[48]={                              //dane pustego instrumentu
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
//DO ZROBIENIA: pusty pattern
zbyte MPTEmptyPattern[2]={0xbf,0xfe};                       //dane pustego patternu
zbyte MPTTableAUDCTLFiltered[4]={0x04,0x02,0x00,0x00};      //tablica AUDCTL dla ktorych maja byc
                                                            //  obslugiwane synthy
zbyte MPTLo16Table[0x40]={                                  //tablica czestotliwosci dla kanalow
  0x00,0xf2,0x33,0x96,                                      //  laczonych (kanal niski)
  0xe2,0x38,0x8c,0x00,
  0x6a,0xe8,0x6a,0xef,
  0x80,0x08,0xae,0x46,
  0xe6,0x95,0x41,0xf6,
  0xb0,0x6e,0x30,0xf6,
  0xbb,0x84,0x52,0x22,
  0xf4,0xc8,0xa0,0x7a,
  0x55,0x34,0x14,0xf5,
  0xd8,0xbd,0xa4,0x8d,
  0x77,0x60,0x4e,0x38,
  0x27,0x15,0x06,0xf7,
  0xe8,0xdb,0xcf,0xc3,
  0xb8,0xac,0xa2,0x9a,
  0x90,0x88,0x7f,0x78,
  0x70,0x6a,0x64,0x5e};

zbyte MPTHi16Table[0x40]={                                  //tablica czestotliwosci dla kanalow
  0x00,0x0d,0x0d,0x0c,                                      //  laczonych (kanal wysoki)
  0x0b,0x0b,0x0a,0x0a,
  0x09,0x08,0x08,0x07,
  0x07,0x07,0x06,0x06,
  0x05,0x05,0x05,0x04,
  0x04,0x04,0x04,0x03,
  0x03,0x03,0x03,0x03,
  0x02,0x02,0x02,0x02,
  0x02,0x02,0x02,0x01,
  0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x00,
  0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00};
zbyte MPTAUDF[4];                                           //wyjsciowy AUDF
zbyte MPTAUDC[4];                                           //wyjsciowy AUDC
zbyte MPTAUDCTL;                                            //wyjsciowy AUDCTL
zbyte MPTChannelAUDCTL[4];                                  //kanalowy AUDCTL
zbyte MPTTempoCounter;                                      //licznik tempa
zbyte MPTTrackPos;                                          //pozycja w tracku
zbyte MPTEmptyPosCounter[4];                                //licznik pustych linii w patternie
zbyte MPTEmptyPosNumber[4];                                 //ilosc pustych linii po nastepnej
                                                            //  nucie w patternie
zword MPTPatternProcPos[4];                                 //pozycja w patternie wg procesora
                                                            //  patternu
zbyte MPTPatternPos;                                        //rzeczywista pozycja w patternie
zbyte MPTJumpCommandEncountered;                            //0xff - nie bylo rozkazu skoku
                                                            //0x00-0x7f - napotkano rozkaz skoku
                                                            //  do tej pozycji w tracku
zbyte MPTChannelNextVolume[4];                              //glosnosc nastepnej nuty
zbyte MPTChannelMainVolume[4];                              //tablica glownych glosnosci kanalow
zbyte MPTChannelMainInstr[4];                               //instrument, ktory bedzie niedlugo
                                                            //  gral na tym kanale
MPTInstrT *(MPTChannelInstrPointer[4]);                     //wskaznik na aktualnie grany
                                                            //  instrument na danym kanale
zint32 MPTChannelInstrPos[4];                               //pozycja kursora w instrumencie
zbyte MPTChannelInstrNote[4];                               //nuta na danym kanale
zbyte MPTChannelInstrFreq[4];                               //czestotliwosc na danym kanale
zbyte MPTChannelInstrVolume[4];                             //aktualna glosnosc na danym kanale
zbyte MPTChannelDecVolCounter[4];                           //licznik zmniejszania sie glosnosci
zbyte MPTChannelTrasposeNumber[4];                          //licznik numerow transpozycji w
                                                            //  instrumencie
zbyte MPTEffectDelay[4];                                    //licznik opoznienia efektu
zbyte MPTChannelEffect[4];                                  //numer efektu na kanale
zbyte MPTChannelEffectParam[4];                             //parametr efektu na kanale
zbyte MPTEffectHelper1[4];                                  //zmienna pomocnicza uzywana przy
                                                            //  efektach
zbyte MPTEffectHelper2[4];                                  //zmienna pomocnicza uzywana przy
                                                            //  efektach
zbyte MPTJoinNote;                                          //numer nuty grany na dzwiekach
                                                            //  dwukanalowych lub 0, jezeli
                                                            //  nie ma dzwieku dwukanalowego
zint32 MPTSongLengthInMs;                                   //dlugosc utworu w milisekundach

/**********************************************************
 funkcje wewnetrzne
 **********************************************************/

#define NOTE_TO_FREQ(channel,note) \
  MPTChannelInstrFreq[channel]=(MPTFreqTable[ \
  ((MPTChannelInstrPointer[channel]->freqtable)&0x30)>>4])[\
  note&0x3f];

//ustawia instrument na odpowiednim kanale
zvoid mpt_SetInstr( zbyte channel,
                    zbyte instr,
                    zbyte note,
                    zbyte volume)
{
  channel=channel&0x03;
  note=note&0x3f;
  volume=volume&0x0f;
  MPTChannelMainVolume[channel]=volume;
  if (instr>=0x20) {
    MPTChannelInstrPointer[channel]=
      (MPTInstrT *)MPTEmptyInstrument;
  } else {
    MPTChannelInstrPointer[channel]=MPTInstrument[instr];
  }
  MPTChannelInstrPos[channel]=0x00000000;
  MPTChannelInstrNote[channel]=note;
  NOTE_TO_FREQ(channel,note)
  MPTChannelInstrVolume[channel]=0x00;
  MPTChannelDecVolCounter[channel]=
    MPTChannelInstrPointer[channel]->decvolumespeed;
  MPTChannelTrasposeNumber[channel]=0;
  MPTEffectDelay[channel]=
    ((MPTChannelInstrPointer[channel]->freqtable)&0x07)<<2;
  MPTChannelEffect[channel]=(MPTChannelInstrPointer[        //znajdywanie numeru efektu
    channel]->effect>>4)&0x07;
  MPTChannelEffectParam[channel]=MPTChannelInstrPointer[    //znajdywanie numeru parametru efektu
    channel]->effect&0x0f;
  MPTEffectHelper1[channel]=0;
  MPTEffectHelper2[channel]=0;
}

//odgrywa instrument na odpowiednim kanale
zvoid mpt_PlayInstr(zbyte channel)
{
  zbyte TempAUDC;
  zbyte TempVolume;
  zbyte TempAccent;
  zbyte TempAccentParam;
  zbyte TempNote;
  zbyte TempEffect;
  zbyte TempEffectParam;
  zbyte TempHelper2;

  channel=channel&0x03;
  //obliczenie aktualnego AUDC
  if (MPTChannelInstrPos[channel]<16) {
    TempAUDC=MPTChannelInstrPointer[channel]->value[
      MPTChannelInstrPos[channel]].audc;
    MPTChannelInstrVolume[channel]=LO_B(TempAUDC);
  } else {
    TempAUDC=MPTChannelInstrPointer[channel]->value[
      15].audc;
    //obsluga opadania glosnosci dzwieku
    if (MPTChannelInstrPointer[channel]->decvolumespeed &&
      MPTChannelInstrVolume[channel]) {
      MPTChannelDecVolCounter[channel]--;
      if (!MPTChannelDecVolCounter[channel]) {
        MPTChannelDecVolCounter[channel]=
          MPTChannelInstrPointer[channel]->decvolumespeed;
        MPTChannelInstrVolume[channel]--;
      }
    }
  }
  //zmniejszanie glosnosci zaleznie od glownej glosnosci
  //  kanalu
  TempVolume=MPTChannelInstrVolume[channel]-0x0f+
    MPTChannelMainVolume[channel];
  if (TempVolume>=0x10) {
    TempVolume=0x00;
  }
  MPTAUDC[channel]=(TempAUDC&0xf0)+TempVolume;

  //obliczenie aktualnego AUDF
  //znajdywanie numeru i parametru akcentu
  TempAccent=0x00;
  TempAccentParam=0x00;
  if (MPTChannelInstrPos[channel]<16) {
    TempAccent=LO_B(MPTChannelInstrPointer[channel]->value[ //znajdywanie numeru akcentu
      MPTChannelInstrPos[channel]].accent);
    if (TempAccent) {
      TempAccentParam=HI_B(MPTChannelInstrPointer[          //znajdywanie numeru parametru
      channel]->value[MPTChannelInstrPos[channel]].accent)
      >>1;
      TempAccentParam=MPTChannelInstrPointer[               //znajdywanie samego parametru
      channel]->AccParam[TempAccentParam];
    }
  }
  //obsluga akcentow
  //  wszystkie akcenty wpisuja 0x00 do AUDCTL chyba ze
  //  pisze inaczej
  MPTChannelAUDCTL[channel]=0x00;
  switch (TempAccent) {
    case 1:                                                 //AUDF<-param
      MPTChannelInstrFreq[channel]=TempAccentParam;
      break;
    case 2:                                                 //AUDF<-AUDF+param
      MPTChannelInstrFreq[channel]+=TempAccentParam;
      break;
    case 3:                                                 //AUDF<-param+numer_nuty
      NOTE_TO_FREQ(channel,MPTChannelInstrNote[channel]+
        TempAccentParam)
      break;
    case 4:                                                 //AUDF<-param i ustawienie zegara
                                                            //  1.79 MHz na kanale (dziala tylko na
                                                            //  kanalach 0 i 2)
      MPTChannelInstrFreq[channel]=TempAccentParam;
      if (channel==0) {
        MPTChannelAUDCTL[channel]=0x40;
      }
      if (channel==2) {
        MPTChannelAUDCTL[channel]=0x20;
      }
      break;
    case 5:                                                 //AUDF<-param i 9-bit poly
      MPTChannelInstrFreq[channel]=TempAccentParam;
      MPTChannelAUDCTL[channel]=0x80;
      break;
    case 6:                                                 //AUDF<-param i ustawienie zegara na
                                                            //  16kHz
      MPTChannelInstrFreq[channel]=TempAccentParam;
      MPTChannelAUDCTL[channel]=0x01;
      break;
    case 7:                                                 //AUDF<-param&random
      MPTChannelInstrFreq[channel]=TempAccentParam&
        (rand()%0xff);
      break;
  }
  if (!TempAccent) {
    //obsluga tremolo
    TempNote=MPTChannelInstrNote[channel];
    TempNote+=MPTChannelInstrPointer[channel]->transpose[
      MPTChannelTrasposeNumber[channel]];
    NOTE_TO_FREQ(channel,TempNote)
    MPTChannelTrasposeNumber[channel]++;
    if (MPTChannelTrasposeNumber[channel]>=4) {
      MPTChannelTrasposeNumber[channel]=0;
    }
    //obsluga efektow
    if (MPTEffectDelay[channel]) {
        --MPTEffectDelay[channel];
    } else {
      TempEffect=MPTChannelEffect[channel];
      TempEffectParam=MPTChannelEffectParam[channel];
      TempHelper2=MPTEffectHelper2[channel];
      //efekt 0: lagodne falowanie
      if (TempEffect==0x00) {
        if (TempHelper2&0x01) {
          if (TempHelper2&0x02) {
            MPTEffectHelper1[channel]=-TempEffectParam;
            MPTChannelInstrFreq[channel]-=TempEffectParam;
          } else {
            MPTEffectHelper1[channel]=TempEffectParam;
            MPTChannelInstrFreq[channel]+=TempEffectParam;
          }
        } else {
          MPTEffectHelper1[channel]=0;
        }
        TempHelper2++;
      }
      //efekt 1: dodanie parametru do AUDF
      if (TempEffect==0x01) {
        MPTEffectHelper1[channel]=-TempEffectParam;
        MPTChannelInstrFreq[channel]-=TempEffectParam;
      }
      //efekt 2: opadanie czestotliwosci
      if (TempEffect==0x02) {
        TempHelper2++;
        MPTEffectHelper1[channel]=TempHelper2;
        MPTChannelInstrFreq[channel]+=TempHelper2;
      }
      //efekt 3: opadanie czestotliwosci po nutach
      if (TempEffect==0x03) {
        TempHelper2++;
        TempNote+=TempHelper2;
        NOTE_TO_FREQ(channel,TempNote)
      }
      //efekt 4: podwyzszanie czestotliwosci
      if (TempEffect==0x04) {
        TempHelper2--;
        MPTEffectHelper1[channel]=TempHelper2;
        MPTChannelInstrFreq[channel]+=TempHelper2;
      }
      //efekt 5: podwyzszanie czestotliwosci po nutach
      if (TempEffect==0x05) {
        TempHelper2--;
        TempNote+=TempHelper2;
        NOTE_TO_FREQ(channel,TempNote)
      }
      //efekt 6: glebokie falowanie
      if (TempEffect==0x06) {
        if (TempEffectParam&0x80) {
          TempHelper2--;
        } else {
          TempHelper2++;
        }
        if (TempEffectParam==TempHelper2) {
          TempEffectParam^=0xff;
        }
        MPTEffectHelper1[channel]=TempHelper2;
        MPTChannelInstrFreq[channel]+=TempHelper2;
      }
      //efekt 7: bardzo glebokie falowanie
      if (TempEffect==0x07) {
        if (TempEffectParam&0x80) {
          TempHelper2--;
        } else {
          TempHelper2++;
        }
        if (TempEffectParam==TempHelper2) {
          TempEffectParam^=0xff;
        }
        TempNote+=TempHelper2;
        NOTE_TO_FREQ(channel,TempNote)
      }
      MPTEffectHelper2[channel]=TempHelper2;
      MPTChannelEffectParam[channel]=TempEffectParam;
    }
    //ustawienie AUDCTL
    MPTChannelAUDCTL[channel]=
      MPTChannelInstrPointer[channel]->audctl;
    //obsluga synthow
    if (MPTChannelAUDCTL[channel]&
      MPTTableAUDCTLFiltered[channel]) {
      MPTAUDF[channel+2]=(MPTFreqTable[
        ((MPTChannelInstrPointer[channel]->freqtable)&0x30)
        >>4])[(TempNote+MPTChannelInstrPointer[
        channel]->AccParam[0])&0x3f]+1;
//ZLE
//      MPTAUDF[channel+2]=(MPTFreqTable[0])[(TempNote+
//        MPTChannelInstrPointer[channel]->AccParam[0])&0xff
//        ]+1;
//
    }
  }
  MPTAUDF[channel]=MPTChannelInstrFreq[channel];
  if (channel==1) {
    MPTJoinNote=0x00;
    if (MPTChannelAUDCTL[channel]&0x10) {
      MPTJoinNote=TempNote&0x3f;
    }
  }
  //ustawienie pozycji na nastepny slupek
  if (MPTChannelInstrPos[channel]<0x40000000) {
    MPTChannelInstrPos[channel]++;
  }
}

zvoid mpt_EvaluateSongForInformation(zvoid)
{
  zbyte EndingTrack;
  zbyte EndingCommand;
  zbyte PatternCommand;
  zbyte EvalTempoCounter;
  zbyte EvalTrackPos;
  zbyte EvalJumpCommandEncountered;
  zbyte EvalEmptyPosCounter[4];
  zbyte EvalEmptyPosNumber[4];
  zint32 EvalPatternProcPos[4];
  zbyte EvalPatternPos;
  zint32 EvalLength;

  //resetowanie wartosci
  for(zbyte i=0;i<4;i++) {
    EvalEmptyPosNumber[i]=0;
    EvalEmptyPosCounter[i]=0;
    EvalPatternProcPos[i]=0;
  }
  EvalTempoCounter=0;
  EvalTrackPos=0;
  EvalPatternPos=0;
  EvalJumpCommandEncountered=0xff;
  EvalLength=0;

  while (1) {
    if (!EvalTempoCounter) {
      //sprawdzenie czy jest koniec songu lub skok
      EndingTrack=4;
      for(zbyte i=0;i<4;i++) {
        if (((MPTTrack[i])[EvalTrackPos].pattern>=0x40) &&
          (EndingTrack==4)) {
          EndingTrack=i;
        }
      }
      if (EndingTrack!=4) {                                 //jezeli natrafiono na koniec songu
                                                            //  lub skok
        EvalJumpCommandEncountered=                         //oblicz gdzie trzeba skoczyc
          (MPTTrack[EndingTrack])[EvalTrackPos].transpose;
        if (EvalJumpCommandEncountered>=0x80) {
          EvalJumpCommandEncountered=0x00;
        }
        if (EvalJumpCommandEncountered==0x00) {             //jezeli to koniec songu to
          break;                                            //powrot ze statusem "koniec songu"
        }
        break;                                              //jezeli nie to powrot ze statusem
                                                            //  "napotkano skok"
      }
      //procesor patternu
      for(zbyte i=3;i<4;i--) {
        if (!EvalEmptyPosCounter[i]) {                      //obsluga pustych pozycji
          //wykonywanie instrukcji z patternu
          EndingCommand=0;
          while (!EndingCommand) {
            PatternCommand=(MPTPattern[(MPTTrack[i])[       //pobranie jednej pozycji z patternu
              EvalTrackPos].pattern])[
              EvalPatternProcPos[i]];
            EvalPatternProcPos[i]++;
            if (PatternCommand<0x40) {                      //obsluga grania nuty
              PatternCommand=0xfe;
            }
            if ((PatternCommand>=0x40) &&                   //obsluga zmiany instrumentu
              (PatternCommand<0x60)) {
            }
            if ((PatternCommand>=0x80) &&                   //obsluga ustawiania ilosci pustych
              (PatternCommand<0xc0)) {                      //  pozycji
              EvalEmptyPosNumber[i]=PatternCommand-0x80;
            }
            if ((PatternCommand>=0xc0) &&                   //obsluga ustawiania glosnosci
              (PatternCommand<0xd0)) {
            }
            if ((PatternCommand>=0xe0) &&                   //obsluga konca patternu
              (PatternCommand<0xf0)) {
              EndingCommand=1;
              EvalPatternPos=MPTPatternLength-1;
            }
            if (PatternCommand>=0xf0) {                     //obsluga wstawiania pustych pozycji
              EvalEmptyPosCounter[i]=EvalEmptyPosNumber[i];
              EndingCommand=1;
            }
          }
        } else {
          EvalEmptyPosCounter[i]--;                         //zmiejszenie licznika pustych pozycji
        }
      }
      //zwiekszenie pozycji w patternie
      EvalPatternPos++;
      //sprawdzenie czy jest koniec patternu
      if (EvalPatternPos>=MPTPatternLength) {
        //zwiekszenie pozycji w songu
        EvalTrackPos++;
        //reset pozycji w patternie
        EvalPatternPos=0;
        for(zbyte i=0;i<4;i++) {
          EvalEmptyPosNumber[i]=0;
          EvalEmptyPosCounter[i]=0;
          EvalPatternProcPos[i]=0;
        }
      }
      EvalTempoCounter=MPTTempo;
    }
    EvalTempoCounter--;
    EvalLength+=20;
  }
  MPTSongLengthInMs=EvalLength;
}

/**********************************************************
 funkcje zewnetrzne
 **********************************************************/

//ladowanie songu

zint32 mpt_LoadFile(zchar * FileName)
{
  zword TempAddr;
  zword TempTrackLength;
  zbyte TempByte;
  zbyte TempPrevByte;

  if ((MPTFile=fopen(FileName,"rb"))==NULL) {return -1;}    //probujemy otwiorzyc plik z
                                                            //  danymi songu
  if ((fgetc(MPTFile)!=0x00ff)||(fgetc(MPTFile)!=0x00ff))   //sprawdzamy 2 pierwsze bajty
  {                                                         //  musza one byc rowne 0xffff
    fclose(MPTFile);
    return -1;
  }
  MPTBegin=fgetc(MPTFile)+(fgetc(MPTFile)<<8);              //czytamy 6 nastepnych bajtow i
  MPTEnd=fgetc(MPTFile)+(fgetc(MPTFile)<<8);                //  ustawiamy kilka zmiennych
  MPTSize=MPTEnd-MPTBegin+1;                                //  potrzebnych przy obliczaniu
                                                            //  adresow instrumentow i patternow
  MPTData=malloc(0x00010000);                               //alokujemy pamiec dla danych songu
  for(zint32 i=0;i<0x01ca;i++) {                            //wczytujemy informacje o songu
    TempByte=(zbyte)(fgetc(MPTFile));
    *(((zbyte *)MPTData)+i)=TempByte;
  }
  TempPrevByte=0x00;
  for(zint32 i=0x01ca;i<MPTSize;i++) {                      //wypelniamy pamiec danymi z pliku
    TempByte=(zbyte)(fgetc(MPTFile));                       //sprawdzamy czy jest 0xffff
    if ((TempByte==0xff) && (TempPrevByte==0xff)) {
      MPTSize+=2;
      TempPrevByte=0x00;
      for(zbyte j=0;j<0x20;j++) {                           //przesuwamy odpowiednie adresy
        TempAddr=(*((zword *)((zbyte *)MPTData+(j<<1))));   //  instrumentow o +2 bajty
        if (TempAddr>=(i-1)) {
          TempAddr+=2;
          *((zword *)((zbyte *)MPTData+(j<<1)))=TempAddr;
        }
      }
      for(zbyte j=0;j<0x40;j++) {                           //przesuwamy odpowiednie adresy
        TempAddr=(*((zword *)((zbyte *)MPTData+(j<<1)+      //  patternow o +2 bajty
          0x40)));
        if (TempAddr>=(i-1)) {
          TempAddr+=2;
          *((zword *)((zbyte *)MPTData+(j<<1)+0x40))=
            TempAddr;
        }
      }
      for(zbyte j=0;j<0x04;j++) {                           //przesuwamy odpowiednie adresy
        TempAddr=*((zbyte *)MPTData+j+0x01c0)+              //  trackow o +2 bajty
          ((*((zbyte *)MPTData+j+0x01c4))<<8);
        if (TempAddr>=(i-1)) {
          TempAddr+=2;
          *((zbyte *)MPTData+j+0x01c0)=TempAddr&0x00ff;
          *((zbyte *)MPTData+j+0x01c4)=(TempAddr&0xff00)>>8;
        }
      }
    } else {
      TempPrevByte=TempByte;
    }
    *(((zbyte *)MPTData)+i)=TempByte;
  }
  fclose(MPTFile);                                          //juz wiecej nie bedziemy potrzebowali
                                                            //  pliku wiec go zamykamy
  for(zbyte i=0;i<0x20;i++) {                               //inicjowanie tablicy wskaznikow do
    TempAddr=(*((zword *)((zbyte *)MPTData+(i<<1))));       //  instrumentow
    if (TempAddr==0x0000) {
      MPTInstrument[i]=(MPTInstrT *)&MPTEmptyInstrument;
    } else {
      MPTInstrument[i]=(MPTInstrT *)((zbyte *)MPTData+
        ((*((zword *)((zbyte *)MPTData+(i<<1))))-MPTBegin));
    }
  }
  for(zbyte i=0;i<0x40;i++) {                               //inicjowanie tablicy wskaznikow do
    TempAddr=(*((zword *)((zbyte *)MPTData+(i<<1)+0x0040)));//  patternow
    if (TempAddr==0x0000) {
      MPTPattern[i]=(zbyte *)&MPTEmptyPattern;
    } else {
      MPTPattern[i]=(zbyte *)MPTData+
        ((*((zword *)((zbyte *)MPTData+(i<<1)+0x0040)))-
        MPTBegin);
    }
  }
  MPTFreqTable[0]=(zbyte *)MPTData+0x00c0;                  //inicjowanie wskaznikow do tablic
  MPTFreqTable[1]=(zbyte *)MPTData+0x0100;                  //  czestotliwosci
  MPTFreqTable[2]=(zbyte *)MPTData+0x0140;
  MPTFreqTable[3]=(zbyte *)MPTData+0x0180;
  for(zbyte i=0;i<4;i++) {                                  //inicjowanie tablicy wskaznikow do
    MPTTrack[i]=(MPTTrackT *)(*((zbyte *)MPTData+0x01c0+i)+ //  trackow
      ((*((zbyte *)MPTData+0x01c4+i))<<8)-MPTBegin+
      (zbyte *)MPTData);
  }
  //obliczenie dlugosci trackow w bajtach
  TempTrackLength=((*((zbyte *)MPTData+0x01c5))<<8)+
    (*((zbyte *)MPTData+0x01c1));
  TempTrackLength-=((*((zbyte *)MPTData+0x01c4))<<8)+
    (*((zbyte *)MPTData+0x01c0));
  //tablice trackow niestety moga byc bledne, chyba blad w
  //  samym zapisie plikow MPT, tutaj je poprawiam
//ZLE
//  for(zbyte i=1;i<4;i++) {
//    MPTTrack[i]=(MPTTrackT *)((zbyte *)MPTTrack[i-1]+TempTrackLength);
//  }
//
  //koniec poprawiania bledow we wskaznikach trackow
  MPTPatternLength=*((zbyte *)MPTData+0x01c8);              //ustawienie dlugosci patternow
  MPTTempo=*((zbyte *)MPTData+0x01c9);                      //ustawienie tempa utworu
  //ustawianie nazwy utworu
  strcpy(MPTSongName,FileName);
  //sprawdzanie czasu trwania utworu
  MPTSongLengthInMs=0x00;
  mpt_EvaluateSongForInformation();
  return 0;
}

//uwalnia pamiec i deinicjalizuje song

zvoid mpt_Free(zvoid)
{
  if (MPTData!=NULL) {                                      //uwalniamy pamiec jezeli song jest
    free(MPTData);                                          //zaladowany
    MPTData=NULL;
  }
}

//resetuje wszystkie zmienne i ustawia player na poczatku
//  odgrywania
zvoid mpt_Reset(zvoid)
{
  for(zbyte i=0;i<4;i++) {
    MPTAUDF[i]=0;
    MPTAUDC[i]=0;
    MPTEmptyPosNumber[i]=0;
    MPTEmptyPosCounter[i]=0;
    MPTPatternProcPos[i]=0;
    MPTChannelNextVolume[i]=0x0f;
    MPTChannelMainInstr[i]=0xff;
    mpt_SetInstr(i,0xff,0x00,0x00);
  }
  MPTAUDCTL=0;
  MPTTempoCounter=0;
  MPTTrackPos=0;
//ZLE
//  MPTTrackPos=0x06;
//
  MPTPatternPos=0;
  MPTJumpCommandEncountered=0xff;
}

//zwraca ilosc wywolan na "ramke" playera
zint32 mpt_GetFPJ(zvoid)
{
  return 1;
}

//zwraca dlugosc songu w milisekundach
zint32 mpt_GetRealSongLengthInMs(zvoid)
{
  return MPTSongLengthInMs;
}

//zint32 tmc_GetRealSongLength(zvoid);

//zwraca wskaznik do nazwy aktualnie zaladowanego utworu
zchar *mpt_GetSongName(zvoid)
{
  return MPTSongName;
}

//pobiera nazwe songu z pliku i wstawia ja do StringTo Fill
zint32 mpt_GetFileInfoName( zchar *FileName,
                            zchar *StringToFill)
{
//  strcpy(StringToFill,FileName);
}

//jedno wywolanie playera
zint32 mpt_PlayerCall(zvoid)
{
  zbyte EndingTrack;
  zbyte EndingCommand;
  zbyte PatternCommand;
  zbyte EndPatternCommandEncountered;

//ZLE
//EvaluateAgain:
//
  EndPatternCommandEncountered=0;
  if (!MPTTempoCounter) {
    //sprawdzenie czy jest koniec songu lub skok
    EndingTrack=4;
    for(zbyte i=0;i<4;i++) {
      if (((MPTTrack[i])[MPTTrackPos].pattern>=0x40) &&
        (EndingTrack==4)) {
        EndingTrack=i;
      }
    }
    if (EndingTrack!=4) {                                   //jezeli natrafiono na koniec songu
                                                            //  lub skok
      MPTJumpCommandEncountered=                            //oblicz gdzie trzeba skoczyc
        (MPTTrack[EndingTrack])[MPTTrackPos].transpose;
      if (MPTJumpCommandEncountered>=0x80) {
        MPTJumpCommandEncountered=0x00;
      }
      if (MPTJumpCommandEncountered==0x00) {                //jezeli to koniec songu to
        return -1;                                          //powrot ze statusem "koniec songu"
      }
      return -2;                                            //jezeli nie to powrot ze statusem
                                                            //  "napotkano skok"
    }
    //procesor patternu
    for(zbyte i=3;i<4;i--) {
      if (!MPTEmptyPosCounter[i]) {                         //obsluga pustych pozycji
        //wykonywanie instrukcji z patternu
        EndingCommand=0;
        while (!EndingCommand) {
          PatternCommand=(MPTPattern[(MPTTrack[i])[         //pobranie jednej pozycji z patternu
            MPTTrackPos].pattern])[MPTPatternProcPos[i]];
          MPTPatternProcPos[i]++;
          if (PatternCommand<0x40) {                        //obsluga grania nuty
            mpt_SetInstr(i,MPTChannelMainInstr[i],(zbyte)(
              PatternCommand+
              (MPTTrack[i])[MPTTrackPos].transpose),
              MPTChannelNextVolume[i]);
            MPTChannelNextVolume[i]=0x0f;
            PatternCommand=0xfe;
          }
          if ((PatternCommand>=0x40) &&                     //obsluga zmiany instrumentu
            (PatternCommand<0x60)) {
            MPTChannelMainInstr[i]=PatternCommand-0x40;
          }
          if ((PatternCommand>=0x80) &&                     //obsluga ustawiania ilosci pustych
            (PatternCommand<0xc0)) {                        //  pozycji
            MPTEmptyPosNumber[i]=PatternCommand-0x80;
          }
          if ((PatternCommand>=0xc0) &&                     //obsluga ustawiania glosnosci
            (PatternCommand<0xd0)) {
            MPTChannelNextVolume[i]=PatternCommand-0xc0;
          }
          if ((PatternCommand>=0xe0) &&                     //obsluga konca patternu
            (PatternCommand<0xf0)) {
//ZLE
//            EndPatternCommandEncountered=1;
//
            EndingCommand=1;
            MPTPatternPos=MPTPatternLength-1;
          }
          if (PatternCommand>=0xf0) {                       //obsluga wstawiania pustych pozycji
            MPTEmptyPosCounter[i]=MPTEmptyPosNumber[i];
            EndingCommand=1;
          }
        }
      } else {
        MPTEmptyPosCounter[i]--;                            //zmiejszenie licznika pustych pozycji
      }
    }
    //zwiekszenie pozycji w patternie
    MPTPatternPos++;
    //sprawdzenie czy jest koniec patternu
    if (MPTPatternPos>=MPTPatternLength) {
      //zwiekszenie pozycji w songu
      MPTTrackPos++;
      //reset pozycji w patternie
      MPTPatternPos=0;
      for(zbyte i=0;i<4;i++) {
        MPTEmptyPosNumber[i]=0;
        MPTEmptyPosCounter[i]=0;
        MPTPatternProcPos[i]=0;
        MPTChannelNextVolume[i]=0x0f;
        MPTChannelMainInstr[i]=0xff;
      }
    }
//ZLE
//    if (EndPatternCommandEncountered) {goto EvaluateAgain;}
//
    MPTTempoCounter=MPTTempo;
  }
  MPTTempoCounter--;
  //odgranie instrumentow
  MPTJoinNote=0x00;
  mpt_PlayInstr(3);
  mpt_PlayInstr(2);
  mpt_PlayInstr(1);
  mpt_PlayInstr(0);
  //obsluga dzwiekow dwukanalowych
  if (MPTJoinNote) {
    MPTAUDF[0]=MPTLo16Table[MPTJoinNote];
    MPTAUDF[1]=MPTHi16Table[MPTJoinNote];
  }
  MPTAUDCTL=MPTChannelAUDCTL[0]|MPTChannelAUDCTL[1]|
    MPTChannelAUDCTL[2]|MPTChannelAUDCTL[3];
}

//zvoid tmc_MakeJump(zvoid);
//zvoid tmc_MakeJumpTo(zbyte NewSongPos,zbyte NewPatternPos);

//skacze z odtwarzaniem songu do miejsca oddalonego od
//  poczatku songu o x milisekund
zvoid mpt_MakeJumpToMs(zint32 ToMs)
{
    int t,tt;

    mpt_Reset();
    t=20;
    tt=0;
    while ((tt+=t)<ToMs) {mpt_PlayerCall();}
}

//zwraca AUDF ktore powinno byc przekazane do pokeya
zbyte mpt_GetAUDF(zbyte channel)
{
  return MPTAUDF[channel&0x03];
}

//zwraca AUDC ktore powinno byc przekazane do pokeya
zbyte mpt_GetAUDC(zbyte channel)
{
  return MPTAUDC[channel&0x03];
}

//zwraca AUDCTL ktore powinno byc przekazane do pokeya
zbyte mpt_GetAUDCTL(zbyte pokey)
{
  return MPTAUDCTL;
}

#endif

