#include <stdlib.h>                                         //instrukcje operacji na pamieci
                                                            //  malloc i free
#include <math.h>                                           //instrukcje matematyczne
#include "pokey.h"                                          //header tej biblioteki

/**********************************************************
 glowne zmienne wystepujace w programie
 **********************************************************/

//zmienne dotyczace odgrywanego dzwieku
zuint32 SampleRate;                                         //czestotliwosc odgrywania sampli w Hz
zint32  SampleCounter;                                      //licznik potrzebny przy obliczaniu
                                                            //  ilosci sampli ktore maja isc do
                                                            //  bufora podczas emulacji pokeya
zuint8  NumberOfPokeys;                                     //ilosc pokeyow branych pod uwage
                                                            //  podczas emulacji
zuint8  NumberOfChannels;                                   //ilosc kanalow branych pod uwage
                                                            //  podczas emulacji
zint32  CyclesCounter;                                      //ilosc cykli do zaemulowania
zfloat  SmoothTable[SMOOTH_TABLE_SIZE];                     //tablica wyokraglajaca dzwiek
zint32  FreqCounterPlayer;                                  //ilosc cykli pozostalych do odegrania
                                                            //  jednego sampla
zuint8  *BufferPointer;                                     //wskaznik na zewnetrzny buffor
                                                            //  do ktorego ida sample
zuint16 BufferSize;                                         //wielkosc tego bufora w ilosci sampli
zuint16 BufferCounter;                                      //pozycja kursora w tym buforze
zuint32 PokeySample[NUMBER_OF_POKEYS];                      //sample z odpowiednich pokeyow ktore
                                                            //  maja trafic do tego bufora
zvoid   (*MixPokeys)(zvoid);                                //wskaznik na funkcje miksujaca dzwiek
zint32  FreqCounterPlayerAdd;                               //zmienna pomocnicza przy obliczaniu
                                                            //  cykli miedzy wywolaniami
                                                            //  wydarzenia na playerze

//zmienne dotyczace kanalow dzwieku
zbyte   AUDF[NUMBER_OF_CHANNELS];                           //tablica czestotliwosci na
                                                            //  odpowiednich kanalach
zbyte   AUDC[NUMBER_OF_CHANNELS];                           //tablica typow dzwiekow
zbyte   AUDCTL[NUMBER_OF_POKEYS];                           //AUDCTL dla pokeyow
zbyte   ChannelState[NUMBER_OF_CHANNELS];                   //stan kanalu (wysoki lub niski)
zint32  FreqCounterChannel[NUMBER_OF_CHANNELS];             //ilosc cykli pozostalych do
                                                            //  "wydarzenia"
zint32  FreqCounterChannelAdd[NUMBER_OF_CHANNELS];          //ilosc cykli dodawanych po wystapieniu
                                                            //  "wydarzenia"
zint32  FreqCounterChannelAddJoin[NUMBER_OF_CHANNELS];      //ilosc cykli dodawanych po wystapieniu
                                                            //  "wydarzenia" (przy kanalach
                                                            //  laczonych)
zint32  FreqCounterChannelPast[NUMBER_OF_CHANNELS];         //ilosc cykli po "wydarzeniu"
zint32  WaveFrom[NUMBER_OF_CHANNELS];                       //wartosc poczatkowa zaokraglenia fali
zint32  WaveTo[NUMBER_OF_CHANNELS];                         //wartosc koncowa zaokraglenia fali

//zmienne dotyczace poly-counters
zuint32 Poly4Value;                                         //tablica wartosci dla 4-bit poly-c
zuint32 Poly5Value;                                         //tablica wartosci dla 5-bit poly-c
zuint32 Poly4Counter[NUMBER_OF_POKEYS];                     //licznik dla 4-bit poly-c
zuint32 Poly5Counter[NUMBER_OF_POKEYS];                     //licznik dla 5-bit poly-c

//pomocnicze tablice AUDCTL
zbyte   AudctlTableHiClock[4]={0x40,0x00,0x20,0x00};        //przy jakich wartosciach AUDCTL
                                                            //  trzeba przestawic zegar kanalu
                                                            //  na 1.79 MHz
zbyte   AudctlTableJoinHi[4]={0x00,0x10,0x00,0x08};         //przy jakich wartosciach AUDCTL
                                                            //  nastepuje laczenie dwoch kanalow
                                                            //  (wartosci dla wyzszego kanalu)
zbyte   AudctlTableJoinLo[4]={0x10,0x00,0x08,0x00};         //przy jakich wartosciach AUDCTL
                                                            //  nastepuje laczenie dwoch kanalow
                                                            //  (wartosci dla nizszego kanalu)
zbyte   AudctlTableFiltering[4]={0x00,0x00,0x04,0x02};      //przy jakich wartosciach AUDCTL
                                                            //  wystepuje hi-pass
                                                            //  (wartosci dla kanalu odmierzajacego)
zbyte   AudctlTableFiltered[4]={0x04,0x02,0x00,0x00};       //przy jakich wartosciach AUDCTL
                                                            //  wystepuje hi-pass
                                                            //  (wartosci dla kanalu
                                                            //  zapamietywanego)

/**********************************************************
 funkcje wewnetrzne
 **********************************************************/

//zwracaja stany poly-counters, 4, 5 i 17 bitowego
//DO ZROBIENIA: obsluga 9-bit poly-counter (rzadko uzywany
//ale jednak)
#define GETPOLY4BIT(pokey) \
  ((Poly4Value&(1<<((Poly4Counter[pokey]+Delay)%15)))&&1)
#define GETPOLY5BIT(pokey) \
  ((Poly5Value&(1<<((Poly5Counter[pokey]+Delay)%31)))&&1)
#define GETPOLY17BIT(pokey) (rand()&0x01)

//zwraca punkt ze srodka tablicy wyokraglajacej na podstawie
//WaveFrom, WaveTo i kanalu
#define WAVEPOINT(channel) \
  (WaveFrom[channel]+(WaveTo[channel]-WaveFrom[channel])* \
  SmoothTable[FreqCounterChannelPast[channel]])

//miksowanie dzwieku dla bufora:

//  16-bit ze znakiem, stereo, dwa pokeye
zvoid MixPokeys_S16_Stereo(zvoid)
{
  zuint8 *TempPointer=BufferPointer+(BufferCounter*4);      //obliczanie pozycji w buforze
                                                            //  do zapisu sampla
  BufferCounter++;                                          //zwiekszenie pozycji kursora
  *((zint16 *)(TempPointer))=                               //zapis sampla w buforze
    (PokeySample[0]+0x80000000)>>16;
  *((zint16 *)(TempPointer+2))=
    (PokeySample[1]+0x80000000)>>16;
}

//  16-bit ze znakiem, mono, jeden pokey
zvoid MixPokeys_S16_Mono(zvoid)
{
  zuint8 *TempPointer=BufferPointer+(BufferCounter*2);      //obliczanie pozycji w buforze
                                                            //  do zapisu sampla
  BufferCounter++;                                          //zwiekszenie pozycji kursora
  *((zint16 *)(TempPointer))=                               //zapis sampla w buforze
    (PokeySample[0]+0x80000000)>>16;
}

//  16-bit ze znakiem, mono, dwa pokeye zmiksowane
zvoid MixPokeys_S16_Mono_Mixed(zvoid)
{
  zuint8 *TempPointer=BufferPointer+(BufferCounter*2);      //obliczanie pozycji w buforze
                                                            //  do zapisu sampla
  BufferCounter++;                                          //zwiekszenie pozycji kursora
  *((zint16 *)(TempPointer))=                               //zapis sampla w buforze
    ((PokeySample[0]>>1)+(PokeySample[1]>>1)+0x80000000)
    >>16;
}

//DO ZROBIENIA: wiecej rodzajow miksowania
//DO ZROBIENIA: rodzaje miksowania z "cichym" poczatkiem i
//  koncem melodii

/**********************************************************
 funkcje zewnetrzne
 **********************************************************/

zvoid InitPokey(zuint32 NewSampleRate,
                zbyte NewNumberOfPokeys)
{
  DonePokey();                                              //deinicjacia emulatora
  SampleRate=NewSampleRate;                                 //zerowanie i inicjacia wszystkich
  SampleCounter=0;                                          //  wartosci
  NumberOfPokeys=NewNumberOfPokeys;
  NumberOfChannels=NumberOfPokeys<<2;
  Poly4Value=0x7ac8;                                        //tablice wartosci dla 4 i 5 bit polys
//ZLE
//  Poly5Value=0x5763e690;
//
  Poly5Value=0x65bd44e0;
  for(int i=0;i<NumberOfPokeys;i++) {
    Poly4Counter[i]=rand()%15;                              //zerowanie licznikow dla 4 i
    Poly5Counter[i]=rand()%31;                              //5 bit polys
    PokeySample[i]=0;                                       //zerowanie sampli
    AUDCTL[i]=0;                                            //i AUDCTL dla pokeyow
  }
  CyclesCounter=0;
  for(zint32 i=0;i<SMOOTH_TABLE_SIZE;i++) {                 //tworzenie tablicy wyokraglajacej
    SmoothTable[i]=zfloat(i)/zfloat(SMOOTH_TABLE_SIZE);
  }
  FreqCounterPlayer=0;
  FreqCounterPlayerAdd=0;
  BufferPointer=NULL;                                       //zerowanie wartosci dla bufora
  BufferSize=0;
  BufferCounter=0;
  MixPokeys=NULL;
  for(zbyte i=0;i<NumberOfChannels;i++) {                   //zerowanie kanalow
    ChannelState[i]=0;
    AUDF[i]=0;
    AUDC[i]=0;
    FreqCounterChannel[i]=0;
    FreqCounterChannelAdd[i]=0;
    FreqCounterChannelAddJoin[i]=0;
    FreqCounterChannelPast[i]=0;
    WaveFrom[i]=0;
    WaveTo[i]=0;
    SoundFC(i,0,0);
  }
  for(zbyte i=0;i<NumberOfChannels;i++) {
    SoundFC(i,0,0);
  }
}

zvoid DonePokey(zvoid)
{
}

zvoid SetCyclesCounter(zint32 cycles)
{
  CyclesCounter=cycles;
}

zvoid SetBufferCounter(zuint16 NewPosition)
{
  BufferCounter=NewPosition;
}

zuint16 GetBufferCounter(zvoid)
{
  return BufferCounter;
}

zvoid SetExternalBuffer(zuint8 *buffer,
                        zuint16 size,
                        zuint8 MixType)
{
  BufferPointer=buffer;
  BufferSize=size;
  BufferCounter=0;
  switch (MixType) {
    case MT_S16_STEREO:
      MixPokeys=MixPokeys_S16_Stereo;
      break;
    case MT_S16_MONO:
      MixPokeys=MixPokeys_S16_Mono;
      break;
    case MT_S16_MONO_MIXED:
      MixPokeys=MixPokeys_S16_Mono_Mixed;
      break;
  }
}

zvoid SoundFC(zbyte channel,
              zbyte NewAUDF,
              zbyte NewAUDC)
{
  zbyte PokeyNumber;
  zint16 FrequencyMultiplier;

  AUDF[channel]=NewAUDF;
  AUDC[channel]=NewAUDC;
  PokeyNumber=channel>>2;
  if (AUDCTL[PokeyNumber]&0x01) {
    FrequencyMultiplier=112;
  } else {
    FrequencyMultiplier=28;
  }
  if (AUDCTL[PokeyNumber]&AudctlTableJoinHi[channel&0x03]) {
    if (AUDCTL[PokeyNumber]&
      AudctlTableHiClock[(channel-1)&0x03]) {
      FrequencyMultiplier=1;
    }
    FreqCounterChannelAdd[channel]=
      (((zint32)(AUDF[channel])<<8)+
      (zint32)(AUDF[channel-1])+1)*
      FrequencyMultiplier;
    FreqCounterChannelAdd[channel-1]=256*
      FrequencyMultiplier;
    FreqCounterChannelAddJoin[channel-1]=
      ((zint32)(AUDF[channel-1])+1)*
      FrequencyMultiplier;
    if (FrequencyMultiplier==1) {
      FreqCounterChannelAdd[channel]+=HI_JOIN_DELAY;
      FreqCounterChannelAddJoin[channel-1]+=HI_JOIN_DELAY;
    }
  } else {
    if (AUDCTL[PokeyNumber]&AudctlTableHiClock[channel&0x03]) {
      FrequencyMultiplier=1;
    }
    if (AUDCTL[PokeyNumber]&AudctlTableJoinLo[channel&0x03]) {
      FreqCounterChannelAdd[channel+1]=
        (((zint32)(AUDF[channel+1])<<8)+
        (zint32)(AUDF[channel])+1)*
        FrequencyMultiplier;
      FreqCounterChannelAdd[channel]=256*
        FrequencyMultiplier;
      FreqCounterChannelAddJoin[channel]=
        ((zint32)(AUDF[channel])+1)*
        FrequencyMultiplier;
      if (FrequencyMultiplier==1) {
        FreqCounterChannelAdd[channel+1]+=HI_JOIN_DELAY;
        FreqCounterChannelAddJoin[channel]+=HI_JOIN_DELAY;
      }
    } else {
      FreqCounterChannelAdd[channel]=
        ((zint32)(AUDF[channel])+1)*FrequencyMultiplier;
      if (FrequencyMultiplier==1) {
        FreqCounterChannelAdd[channel]+=HI_NOJOIN_DELAY;
      }
    }
  }
  if (AUDC[channel]&0x10) {
    if (WaveTo[channel]!=((AUDC[channel]&0x0f)<<26)) {
      if (FreqCounterChannelPast[channel]<SMOOTH_TABLE_SIZE) {
        WaveFrom[channel]=(zint32)(WAVEPOINT(channel));
      } else {
        WaveFrom[channel]=WaveTo[channel];
      }
      WaveTo[channel]=(AUDC[channel]&0x0f)<<26;
      FreqCounterChannelPast[channel]=0;
    }
  }
}

zvoid SetAUDCTL(zbyte pokey,zbyte NewAUDCTL)
{
  AUDCTL[pokey]=NewAUDCTL;
}

zuint8 EmulatePokeys(zvoid)
{
  zuint8 PokeyNumber,ChangeState;
  zint32 MinimalCycles;                                     //cykle pozostale do jakiegos
                                                            //  wydarzenia
  zuint8 MinimalCyclesType;                                 //typ nastepnego wydarzenia
                                                            //0x00-0x20 - wydarzenie na kanale
                                                            //  o tym numerze
                                                            //0xfe - wydarzenie na playerze
                                                            //0xff - koniec cykli
  zuint8 Delay;

  //glowna petla
  while (1) {
    //znajdowanie najblizszego wydarzenia
    MinimalCycles=CyclesCounter;                            //badanie najpierw konca cykli
    MinimalCyclesType=0xff;
    if (MinimalCycles>=FreqCounterPlayer) {                 //pozniej wydarzenia na playerze
      MinimalCycles=FreqCounterPlayer;
      MinimalCyclesType=0xfe;
    }
    for(zint8 i=NumberOfChannels-1;i>=0;i--) {              //i na koniec badanie kanalow
      if (MinimalCycles>=FreqCounterChannel[i]) {
        MinimalCycles=FreqCounterChannel[i];
        MinimalCyclesType=i;
      }
    }
    //przesuniecie wszystkich licznikow do najblizszego
    //wydarzenia
    for(zuint8 i=0;i<NumberOfChannels;i++) {
      FreqCounterChannel[i]-=MinimalCycles;                 //wszystkie liczniki wydarzen kanalow
      FreqCounterChannelPast[i]+=MinimalCycles;             //i wszystkie liczniki kanalow
                                                            //"po wydarzeniu"
    }
    FreqCounterPlayer-=MinimalCycles;                       //licznik wydarzenia na playerze
    CyclesCounter-=MinimalCycles;
    //przesuniecie licznikow poly-counters
    for(int i=0;i<NumberOfPokeys;i++) {
      Poly4Counter[i]=(Poly4Counter[i]+MinimalCycles)       //4-bitowych
        %15;
      Poly5Counter[i]=(Poly5Counter[i]+MinimalCycles)       //i 5-bitowych
        %31;
    }
    //obsluga wydarzenia "zabraklo cykli"
    if (MinimalCyclesType==0xff) {return 1;}                //powrot z funkcji ze statusem
                                                            //  "zabraklo cykli"
    //obsluga wydarzenia "wydarzenie na kanale x"
    if (MinimalCyclesType<0x20) {
      PokeyNumber=MinimalCyclesType>>2;                     //obliczanie numeru pokeya do ktorego
                                                            //  nalezy kanal w ktorytm wystapilo
                                                            //  wydarzenie
      Delay=MinimalCyclesType&0x03;
      switch (AUDC[MinimalCyclesType]&0xf0) {               //badanie czy ma sie zmienic stan
        case 0x00:                                          //  kanalu (wysoki/niski)
          ChangeState=GETPOLY5BIT(PokeyNumber)&             //  na podstawie AUDC kanalu
            GETPOLY17BIT(PokeyNumber);
          break;
        case 0x20:
          ChangeState=GETPOLY5BIT(PokeyNumber);
          break;
        case 0x40:
          ChangeState=GETPOLY5BIT(PokeyNumber)&
            GETPOLY4BIT(PokeyNumber);
          break;
        case 0x60:
          ChangeState=GETPOLY5BIT(PokeyNumber);
          break;
        case 0x80:
          ChangeState=GETPOLY17BIT(PokeyNumber);
          break;
        case 0xa0:
          ChangeState=0x01;
          break;
        case 0xc0:
          ChangeState=GETPOLY4BIT(PokeyNumber);
          break;
        case 0xe0:
          ChangeState=0x01;
          break;
        default:
          ChangeState=0x00;
      }
      //obliczanie wygladu fali po zmianie stanu
      //kanalu
      if (ChangeState) {
        if (FreqCounterChannelPast[
          MinimalCyclesType]<SMOOTH_TABLE_SIZE) {
          WaveFrom[MinimalCyclesType]=(zint32)(
            WAVEPOINT(MinimalCyclesType));
        } else {
          WaveFrom[MinimalCyclesType]=
            WaveTo[MinimalCyclesType];
        }
        FreqCounterChannelPast[MinimalCyclesType]=0;
        ChannelState[MinimalCyclesType]^=0x01;
        if (ChannelState[MinimalCyclesType]) {
          WaveTo[MinimalCyclesType]=
            (AUDC[MinimalCyclesType]&0x0f)<<26;
        } else {
          WaveTo[MinimalCyclesType]=0x00;
        }
      }
      //obsluga hi-pass filters
      if (AUDCTL[PokeyNumber]&AudctlTableFiltering[
        MinimalCyclesType&0x03]) {
        if (ChannelState[MinimalCyclesType-2]) {            //zmiana stanu kanalu filtrowanego
          if (FreqCounterChannelPast[                       //jezeli ten stan jest wysoki
            MinimalCyclesType-2]<
            SMOOTH_TABLE_SIZE) {
            WaveFrom[MinimalCyclesType-2]=(zint32)(
              WAVEPOINT(MinimalCyclesType-2));
          } else {
            WaveFrom[MinimalCyclesType]=
              WaveTo[MinimalCyclesType];
          }
          FreqCounterChannelPast[
            MinimalCyclesType-2]=0;
          ChannelState[MinimalCyclesType-2]^=0x01;
          if (ChannelState[MinimalCyclesType-2]) {
            WaveTo[MinimalCyclesType-2]=
              (AUDC[MinimalCyclesType-2]&0x0f<<26);
          } else {
            WaveTo[MinimalCyclesType-2]=0x00;
          }
        }
      }
      //przeliczenie wystapienia nastepnego wydarzenia
      //dla tego kanalu
      FreqCounterChannel[MinimalCyclesType]=
        FreqCounterChannelAdd[MinimalCyclesType];
      if (AUDCTL[PokeyNumber]&AudctlTableJoinHi[            //przeliczanie przy kanalach laczonych
        MinimalCyclesType&0x03]) {
        FreqCounterChannel[MinimalCyclesType-1]=
          FreqCounterChannelAddJoin[
          MinimalCyclesType-1];
      }
      //koniec obslugi wydarzenia
      //"wydarzenie na kanale x"
    }
    //obsluga wydarzenia "wydarzenie na playerze"
    if (MinimalCyclesType==0xfe) {
      if (BufferCounter>=BufferSize) {return 2;}            //wyjscie z funkcji gdy skonczy sie
                                                            //  miejsce w buforze
      for(zbyte i=0;i<NumberOfPokeys;i++) {                 //zerowanie sampli dla pokeyow
        PokeySample[i]=0;
      }
      for(zbyte i=0;i<NumberOfChannels;i++) {               //obliczanie sampli dla poszczegolnych
        PokeyNumber=i>>2;                                   //  pokeyow
//ZLE
//if (i==1) {
//
        if (FreqCounterChannelPast[i]<
          SMOOTH_TABLE_SIZE) {
          PokeySample[PokeyNumber]+=(zint32)(WAVEPOINT(i));
        } else {
          PokeySample[PokeyNumber]+=WaveTo[i];
        }
//ZLE
//}
//
      }
      MixPokeys();                                          //mixowanie sampli z pokeyow i
                                                            //zapisywanie ich w buforze
      FreqCounterPlayer=0;
      while (FreqCounterPlayerAdd<HIGH_CLOCK) {
        FreqCounterPlayer++;
        FreqCounterPlayerAdd+=SampleRate;
      }
      FreqCounterPlayerAdd-=HIGH_CLOCK;
/*      FreqCounterPlayer=FreqCounterPlayerAdd;*/               //obliczenie nastepnego wystapienia
                                                            //wydarzenia w pokeyu
    }
  }
}

