/**********************************************************
 this is emulation liblary for pokey
 written by adept_zap
 still needs a lot of work

 there is a 8 channels in that emulator
  chanels 0-3 for left pokey
  chanels 4-7 for right pokey
 in stereo mode:
  output from left pokey goes to left speaker and
  output from right pokey goes to right speaker
 in mono mode:
  output from both pokeys will be mixed

 internal buffer is always 256 samples wide
 if you will make internal buffer overflow you will loose
 data from the beginning of the buffer
 **********************************************************/

#ifndef POKEY_H
#define POKEY_H

#include "ztypes.h"

#define POKEY_VER   0.95                                    //file version
#define HIGH_CLOCK  1773447                                 //freq of high speed clock (in Hz)
#define MID_CLOCK   63337                                   //freq of middle speed clock (in Hz)
#define LOW_CLOCK   15834                                   //freq of low speed clock (in Hz)
#define NUMBER_OF_POKEYS    2                               //number of pokeys in liblary
#define NUMBER_OF_CHANNELS  (0x04*NUMBER_OF_POKEYS)         //number of channels in liblary
#define SMOOTH_TABLE_SIZE 40                                //size of table used to smooth sound
#define HI_NOJOIN_DELAY 3                                   //delay in cycles for hi_clock sounds
                                                            //  if channels are not joined
#define HI_JOIN_DELAY 6                                     //delay in cycles for hi_clock sounds
                                                            //  if channels are joined

/**********************************************************
 mixing types
 constants gived to SetExternalBuffor function
 **********************************************************/
#define MT_S16_STEREO     0                                 //4-bytes sample, 16-bit signed stereo
                                                            //  2 pokeys, first for left speaker
                                                            //  second for right channel
#define MT_S16_MONO       1                                 //2-bytes sample, 16-bit signed mono
                                                            //  1 pokey for all speakers
#define MT_S16_MONO_MIXED 2                                 //2-bytes sample, 16-bit signed mono
                                                            //  2 pokeys mixed together for
                                                            //  all speakers

/**********************************************************
 initialises pokey library, resets all internal values
 arguments:
  NewSampleRate - speed of the output stream (between
    8000 and 192000), 44100 is most common value
  NewNumberOfPokeys - active pokeys (sounds from that
    pokeys will be computed)
 **********************************************************/
zvoid InitPokey(zuint32 NewSampleRate,
                zbyte NewNumberOfPokeys);

/**********************************************************
 deinitializes pokeys
 **********************************************************/
zvoid DonePokey(zvoid);

/**********************************************************
 sets internal cycle counter, from that value EmulatePokeys
 knows how long it should play before exiting
 arguments:
  NewCyclesCounter - ammount of cycles to play
 **********************************************************/
zvoid SetCyclesCounter(zint32 NewCyclesCounter);

/**********************************************************
 sets position in buffer, from that value EmulatePokeys
 knows when buffer will be full.
 mostly used to clear buffer after it was send to output
 device
 arguments:
  NewCyclesCounter - ammount of cycles to play
 **********************************************************/
zvoid SetBufferCounter(zuint16 NewPosition);

/**********************************************************
 returns position in buffer, from that value EmulatePokeys
 knows when buffer will be full
 **********************************************************/
zuint16 GetBufferCounter(zvoid);

/**********************************************************
 sets external buffer values
 arguments:
  NewBufferPointer - pointer to external buffer
  NewBufferSize - size of external buffer (in samples)
  MixType - way of sound mixing and sample size, see
    "mixing types"
 **********************************************************/
zvoid SetExternalBuffer(zuint8 *NewBufferPointer,
                        zuint16 NewBufferSize,
                        zuint8 MixType);

/**********************************************************
 sets certain sound type on certain channel
 it differs from Sound function only that it takes three
 parameters instead of four
 arguments:
  channel - sound will be changed on this channel (0-7)
  NewAUDF - sound frequency (0-255) (AUDF)
  NewAUDC - sound type (0-255) (AUDC)
    higher 3-bit part of this value is sound distortion
    lower 4-bit part is sound volume
    bit 4, that in the middle is not working yet
 **********************************************************/
zvoid SoundFC(zbyte channel,
              zbyte NewAUDF,
              zbyte NewAUDC);

/**********************************************************
 sets behavior of certain pokey
 arguments:
  NewAUDCTL - bit 0 - not working yet
              bit 1 - inserts high-pass filter into
                      channel 1, clocked by channel 3
              bit 2 - inserts high-pass filter into
                      channel 0, clocked by channel 2
              bits 3-7 - not working yet
 **********************************************************/
zvoid SetAUDCTL(zbyte pokey,zbyte NewAUDCTL);

/**********************************************************
 this will rotate poly-counters and fills buffer with
 samples
 return values:
  it returns 1 - when all cycles was emulated
  or returns 2 - whan buffer is full
 **********************************************************/
zuint8 EmulatePokeys(zvoid);

#endif                                                      //end of POKEY_H

