#include <stdlib.h>                                         //instrukcje operacji na pamieci
                                                            //  malloc i free
#include <stdio.h>                                          //zmienna FILE i instrukcje operacji
                                                            //  na plikach
#include "tmc_player.h"                                     //header tej biblioteki

#ifdef TMC_COMPILE

/**********************************************************
 wykorzystywane struktury
 **********************************************************/

struct SongInfoT{                                           //struktura informacji o melodii
  zchar SongName[0x1e];                                     //  nazwa melodii
  zbyte Tempo;                                              //  tempo melodii
  zbyte FPJ;                                                //  ilosc wywolan playera na ramke
                                                            //    czyli 50*FPJ bps
  zbyte LoInstrAddr[0x40];                                  //  niskie bajty adresow instrumentow
  zbyte HiInstrAddr[0x40];                                  //  wysokie bajty adresow instrumentow
  zbyte LoPattAddr[0x80];                                   //  niskie bajty adresow patternow
  zbyte HiPattAddr[0x80];                                   //  wysokie bajty adresow patternow
};

struct SongPosT{                                            //struktura informacji w danych songu
  zbyte t;                                                  //  transpozycja
  zbyte p;                                                  //  odgrywany pattern
};

struct InstrPosT{                                           //struktura danych jednego slupku
                                                            //    instrumentu
  zbyte primary;                                            //  glosnosc glowna i brzmienie
  zbyte secondary;                                          //  glosnosc pomocnicza i efekt
  zbyte argument;                                           //  argument efektu
};

struct InstrT{                                              //struktura danych instrumentu
  InstrPosT value[21];                                      //  21 slupkow
  zbyte Transpos[8];                                        //  8 transpozycji
  zbyte Param[9];                                           //  9 parametrow instrumentu
};

/**********************************************************
 tablice pomocnicze
 **********************************************************/

zbyte FreqTable[0x0100]={                                   //tablica zamiany nuty na czestotliwosc
  0x00,0xf1,0xe4,0xd7,                                      //  a dokladniej 4 takie tablice po 64
  0xcb,0xc0,0xb5,0xaa,                                      //  bajty kazda
  0xa1,0x98,0x8f,0x87,
  0x7f,0x78,0x72,0x6b,
  0x65,0x5f,0x5a,0x55,
  0x50,0x4b,0x47,0x43,
  0x3f,0x3c,0x38,0x35,
  0x32,0x2f,0x2c,0x2a,
  0x27,0x25,0x23,0x21,
  0x1f,0x1d,0x1c,0x1a,
  0x18,0x17,0x16,0x14,
  0x13,0x12,0x11,0x10,
  0x0f,0x0e,0x0d,0x0c,
  0x0b,0x0a,0x09,0x08,
  0x07,0x06,0x05,0x04,
  0x03,0x02,0x01,0x00,

  0x00,0xf2,0xe6,0xda,
  0xce,0xbf,0xb6,0xaa,
  0xa1,0x98,0x8f,0x89,
  0x80,0x7a,0x71,0x6b,
  0x65,0x5f,0x5c,0x56,
  0x50,0x4d,0x47,0x44,
  0x3e,0x3c,0x38,0x35,
  0x32,0x2f,0x2d,0x2a,
  0x28,0x25,0x23,0x21,
  0x1f,0x1d,0x1c,0x1a,
  0x18,0x17,0x16,0x14,
  0x13,0x12,0x11,0x10,
  0x0f,0x0e,0x0d,0x0c,
  0x0b,0x0a,0x09,0x08,
  0x07,0x06,0x05,0x04,
  0x03,0x02,0x01,0x00,

  0x00,0xff,0xf1,0xe4,
  0xd8,0xca,0xc0,0xb5,
  0xab,0xa2,0x99,0x8e,
  0x87,0x7f,0x79,0x73,
  0x70,0x66,0x61,0x5a,
  0x55,0x52,0x4b,0x48,
  0x43,0x3f,0x3c,0x39,
  0x37,0x33,0x30,0x2d,
  0x2a,0x28,0x25,0x24,
  0x21,0x1f,0x1e,0x1c,
  0x1b,0x19,0x17,0x16,
  0x15,0x13,0x12,0x11,
  0x10,0x0f,0x0e,0x0d,
  0x0c,0x0b,0x0a,0x09,
  0x08,0x07,0x06,0x05,
  0x04,0x03,0x02,0x01,

  0x00,0xf3,0xe6,0xd9,
  0xcc,0xc1,0xb5,0xad,
  0xa2,0x99,0x90,0x88,
  0x80,0x79,0x72,0x6c,
  0x66,0x60,0x5b,0x55,
  0x51,0x4c,0x48,0x44,
  0x40,0x3c,0x39,0x35,
  0x32,0x2f,0x2d,0x2a,
  0x28,0x25,0x23,0x21,
  0x1f,0x1d,0x1c,0x1a,
  0x18,0x17,0x16,0x14,
  0x13,0x12,0x11,0x10,
  0x0f,0x0e,0x0d,0x0c,
  0x0b,0x0a,0x09,0x08,
  0x07,0x06,0x05,0x04,
  0x03,0x02,0x01,0x00};

zbyte Lo16Table[0x40]={                                     //tablica czestotliwosci dla kanalow
  0x00,0xf2,0x33,0x96,                                      //  laczonych (kanal niski)
  0xe2,0x38,0x8c,0x00,
  0x6a,0xe8,0x6a,0xef,
  0x80,0x08,0xae,0x46,
  0xe6,0x95,0x41,0xf6,
  0xb0,0x6e,0x30,0xf6,
  0xbb,0x84,0x52,0x22,
  0xf4,0xc8,0xa0,0x7a,
  0x55,0x34,0x14,0xf5,
  0xd8,0xbd,0xa4,0x8d,
  0x77,0x60,0x4e,0x38,
  0x27,0x15,0x06,0xf7,
  0xe8,0xdb,0xcf,0xc3,
  0xb8,0xac,0xa2,0x9a,
  0x90,0x88,0x7f,0x78,
  0x70,0x6a,0x64,0x5e};

zbyte Hi16Table[0x40]={                                     //tablica czestotliwosci dla kanalow
  0x00,0x0d,0x0d,0x0c,                                      //  laczonych (kanal wysoki)
  0x0b,0x0b,0x0a,0x0a,
  0x09,0x08,0x08,0x07,
  0x07,0x07,0x06,0x06,
  0x05,0x05,0x05,0x04,
  0x04,0x04,0x04,0x03,
  0x03,0x03,0x03,0x03,
  0x02,0x02,0x02,0x02,
  0x02,0x02,0x02,0x01,
  0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x00,
  0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00};

zbyte VolTable[0x10][0x10]={                                //tablica glosnosci
  {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00},
  {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x01,
    0x01,0x01,0x01,0x01,0x01},
  {0x00,0x00,0x00,0x00,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
    0x01,0x02,0x02,0x02,0x02},
  {0x00,0x00,0x00,0x01,0x01,0x01,0x01,0x01,0x02,0x02,0x02,
    0x02,0x02,0x03,0x03,0x03},
  {0x00,0x00,0x01,0x01,0x01,0x01,0x02,0x02,0x02,0x02,0x03,
    0x03,0x03,0x03,0x04,0x04},
  {0x00,0x00,0x01,0x01,0x01,0x02,0x02,0x02,0x03,0x03,0x03,
    0x04,0x04,0x04,0x05,0x05},
  {0x00,0x00,0x01,0x01,0x02,0x02,0x02,0x03,0x03,0x04,0x04,
    0x04,0x05,0x05,0x06,0x06},
  {0x00,0x00,0x01,0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x05,
    0x05,0x06,0x06,0x07,0x07},
  {0x00,0x01,0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x05,0x05,
    0x06,0x06,0x07,0x07,0x08},
  {0x00,0x01,0x01,0x02,0x02,0x03,0x04,0x04,0x05,0x05,0x06,
    0x07,0x07,0x08,0x08,0x09},
  {0x00,0x01,0x01,0x02,0x03,0x03,0x04,0x05,0x05,0x06,0x07,
    0x07,0x08,0x09,0x09,0x0a},
  {0x00,0x01,0x01,0x02,0x03,0x04,0x04,0x05,0x06,0x07,0x07,
    0x08,0x09,0x0a,0x0a,0x0b},
  {0x00,0x01,0x02,0x02,0x03,0x04,0x05,0x06,0x06,0x07,0x08,
    0x09,0x0a,0x0a,0x0b,0x0c},
  {0x00,0x01,0x02,0x03,0x03,0x04,0x05,0x06,0x07,0x08,0x09,
    0x0a,0x0a,0x0b,0x0c,0x0d},
  {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x07,0x08,0x09,
    0x0a,0x0b,0x0c,0x0d,0x0e},
  {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,
    0x0b,0x0c,0x0d,0x0e,0x0f}
};

zbyte InversedChannel[8]={4,5,6,7,0,1,2,3};                 //numery kanalow odwrotnych

/**********************************************************
 zmienne globalne
 **********************************************************/

//zmienne dotyczace bezposrednio pliku

FILE    *TMCFile;                                           //aktualnie odtwarzany plik TMC
zint32  TMCBegin;                                           //poczatek pliku w formacie Atari
zint32  TMCEnd;                                             //koniec pliku w formacie Atari
zint32  TMCSize;                                            //wielkosc pliku w formacie Atari
zvoid   *TMCData=NULL;                                      //adres do poczatku danych pliku
struct  SongInfoT *SongInfo;                                //adres do informacji o pliku
struct  SongPosT (*SongData)[0x80][0x08];                   //dane songu (odgrywane
                                                            //patterny i transpozycje)
struct  InstrT *(Instrument[0x40]);                         //adresy do danych instrumentow
zbyte   (*(Pattern[0x80]))[0x0101];                         //adresy do danych patternow
zbyte   EmptyPattern=0xff;                                  //pusty pattern

//zmienne dotyczace odtwarzania instrumentow

zint32  InstrCounter[8];                                    //pozycja w instrumencie
                                                            //  podczas jego odtwarzania
zint32  InstrNum[8];                                        //numer instrumentu aktywnego na
                                                            //  danym kanale
zbyte   InstrTune[8];                                       //ton podstawowy instrumentu na
                                                            //  danym kanale
zbyte   ActualTune[8];                                      //ton aktualny instrumentu na
                                                            //  danym kanale
zint32  ToTranspose[8];                                     //ilosc wywolan playera pozostalych do
                                                            //  transpozycji tonu w instrumencie
zint32  TransposeNum[8];                                    //numer aktualniej transpozycji w
                                                            //  instrumencie
zbyte   ChannelTranspose[8];                                //transpozycja na calym kanale
zint32  PriToVolumeDec[8];                                  //ilosc wywolan playera pozostalych do
                                                            //  zmniejszenia glosnosci glownej
zint32  SecToVolumeDec[8];                                  //ilosc wywolan playera pozostalych do
                                                            //  zmniejszenia glosnosci dodatkowej
zbyte   AUDFV[8];                                           //aktualna wartosc AUDF majaca byc
                                                            //  przeslana do pokeya
zbyte   AUDCV[8];                                           //aktualna wartosc AUDC majaca byc
                                                            //  przeslana do pokeya
zbyte   ChannelFreq[8];                                     //czestotliwosc dzwieku na kanale
zbyte   ChannelDist[8];                                     //znieksztalcenie kanalu, wartosc
                                                            //  przesunieta w lewo o 4 bity
zbyte   FreqTableToUse[8];                                  //numer uzywanej na kanale tablicy
                                                            //  czestotliwosci przesuniety o
                                                            //  6 bitow w lewo
zbyte   PriVolume[8];                                       //glosnoscglowna kanalu
zbyte   SecVolume[8];                                       //glosnosc dodatkowa kanalu
zbyte   ChannelPlays[8];                                    //1-gdy na kanale gra jakis instrument

#ifdef COMPILE_HIGH_PASS_FILTERS
zbyte   Audctl[8];                                          //aktualna wartosc AUDCTL kanalu
zbyte   AUDCTLV[2]={0,0};                                   //wyjsciowe wartosci AUDCTL
#endif

#ifdef COMPILE_EFFECTS
zbyte   DelayToEffect[8];                                   //opoznienie po ktorym efekt zaczyna
                                                            //  grac
zbyte   EffectHelper1[8];                                   //zmienna pomocnicza uzywana przy
                                                            //  obliczaniu efektow
zbyte   EffectHelper2[8];                                   //zmienna pomocnicza uzywana przy
                                                            //  obliczaniu efektow
zbyte   EffectParam[8];                                     //parametr efektu
zbyte   EffectTempo[8];                                     //szybkosc zmian w efekcie
zbyte   EffectToTempo[8];                                   //licznik tempa w efekcie
#endif

//zmienne dotyczace procesora patternu

zint32  PosInPattern[8];                                    //pozycja w patternie
                                                            //  (wg procesora patternu)
zint32  EmptyPos[8];                                        //ilosc pozostalych pustych miejsc
                                                            //  do miejsca gdzie bedzie wiecej
                                                            //  danych dla procesora patternu
zbyte   SubVolume[8];                                       //najwieksza glosnosc kanalu ustalana
                                                            //  przez pattern (eor 0xff)

//zmnienne dotyczace danych songu

zint32  ToTempo;                                            //ilosc wywolan playera pozostalych do
                                                            //odczytania kolejnej pozycji w
                                                            //patternie
zint32  Tempo;                                              //tempo odtwarzania songu
zint32  PatternPos;                                         //pozycja kursora w patternie
zint32  SongLength;                                         //dlugosc songu w ilosci wpisow w
                                                            //danych songu
zint32  RealSongLength;                                     //dlugosc songu do napotkania
                                                            //pierwszej instrukcji skoku w songu
                                                            //w ilosci wpisow
zint32  RealSongLengthInMs;                                 //dlugosc songu do napotkania
                                                            //pierwszej instrukcji skoku w songu
                                                            //w ilosci wpisow
zint32  SongPos;                                            //pozycja kursora w danych songu
zbyte   SongName[0x1f]="none";                              //nazwa songu, konczona \0
struct  InstrT EmptyInstrument={
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0};

/**********************************************************
 funkcje wewnetrzne
 **********************************************************/

zvoid tmc_EvaluateSongForInformation(zvoid)
{
    zint32 emptyp[8];
    zint32 posinp[8];

    zint32 totmp=0;
    zint32 leavepp=0;
    zint32 songp=0;
    zint32 pattpos=0;
    zint32 tmp=0;
    zint32 elapsed=0;
    zbyte t,tt,ttt;

    tmp=SongInfo->FPJ*((SongInfo->Tempo)+1);
    for(int i=0;i<8;i++) {
        emptyp[i]=0;
        posinp[i]=0;
    }
    elapsed=0;
    while (1) {
        if (totmp==0) {
            ttt=0;
            for(int j=0;j<8;j++) {
                ttt=ttt|((*SongData)[songp][j].p);
            }
            if (ttt&0x80) {
                RealSongLengthInMs=(zint32)(elapsed*20.0/SongInfo->FPJ);
                return;
            }
            if (songp>=SongLength) {
                RealSongLengthInMs=(zint32)(elapsed*20.0/SongInfo->FPJ);
                return;
            }
            for(int i=7;i>=0;i--) {
                leavepp=0;
                if (emptyp[i]>0x00) {
                    emptyp[i]--;
                    leavepp=1;
                }
                while (!leavepp) {
                    t=(*(Pattern[(*SongData)[songp][7-i].p]))[posinp[i]];
                    if (t==0x00) {
                        posinp[i]+=2;
                        leavepp=1;
                    }
                    if ((t>=0x01)&&(t<=0x3f)) {
                        posinp[i]+=2;
                        leavepp=1;
                    }
                    if (t==0x40) {
                        posinp[i]++;
                        tt=(*(Pattern[(*SongData)[songp][7-i].p]))[posinp[i]];
                        if (tt&0x7f) {
                            tmp=((tt&0x7f)+1)*SongInfo->FPJ;
                            totmp=0;
                        } else {
                            pattpos=0x40;
                        }
                        if (tt&0x80) {
                            posinp[i]++;
                        }
                        posinp[i]++;
                        leavepp=1;
                    }
                    if ((t>=0x41)&&(t<=0x7f)) {
                        posinp[i]++;
                        tt=(*(Pattern[(*SongData)[songp][7-i].p]))[posinp[i]];
                        if (tt&0x7f) {
                            tmp=((tt&0x7f)+1)*SongInfo->FPJ;
                            totmp=0;
                        } else {
                            pattpos=0x40;
                        }
                        posinp[i]+=2;
                        leavepp=1;
                    }
                    if ((t>=0x80)&&(t<=0xbf)) {
                        posinp[i]++;
                    }
                    if ((t>=0xc0)&&(t<=0xff)) {
                        emptyp[i]=t&0x3f;
                        posinp[i]++;
                        leavepp=1;
                    }
                }
            }
        }
        if (totmp==0) {
            pattpos++;
            if (pattpos>=0x40) {
                pattpos=0x00;
                for(int i=0;i<8;i++) {
                    emptyp[i]=0x00;
                    posinp[i]=0;
                }
                songp++;
            }
        }
        totmp++;
        if (totmp>=tmp) {totmp=0;}
        elapsed++;
    }
}

//zamienia ton na czestotliwosc

zbyte tmc_TuneToFreq(zbyte ftune)
{
    return (FreqTable[ftune]);
}

//zmienia aktualny instrument na odpowiednim kanale
//  channel-kanal,
//  NewInstrNum-numer instrumentu,

zvoid tmc_ChangeInstrument(zbyte channel,zbyte NewInstrNum)
{
    channel=channel&0x07;
    InstrNum[channel]=NewInstrNum&0x3f;
}

//zaczyna odgrywac dany instrument
//  zeruje jego zmienne
//  channel-kanal,
//  NewTune-numer nuty

zvoid tmc_PlayInstrument(zbyte channel,zbyte NewTune,zbyte NewChannelTranspose)
{
    channel=channel&0x07;
    InstrTune[channel]=NewTune&0x3f;
    InstrCounter[channel]=0;
    PriToVolumeDec[channel]=0;
    SecToVolumeDec[channel]=0;
    ToTranspose[channel]=0;
    TransposeNum[channel]=0;
    ChannelTranspose[channel]=NewChannelTranspose;
    FreqTableToUse[channel]=
        (Instrument[InstrNum[channel]]->Param[7])&0xc0;
    ActualTune[channel]=InstrTune[channel]
        +FreqTableToUse[channel]+ChannelTranspose[channel];
    ChannelFreq[channel]=tmc_TuneToFreq(ActualTune[channel]);
    PriVolume[channel]=0;
    SecVolume[channel]=0;
    ChannelPlays[channel]=1;
#ifdef COMPILE_EFFECTS
    DelayToEffect[channel]=Instrument[InstrNum[channel]]->Param[6];
    EffectHelper1[channel]=0;
    EffectHelper2[channel]=0;
    EffectParam[channel]=(Instrument[InstrNum[channel]]->Param[5])&0x0f;
    if ((Instrument[InstrNum[channel]]->Param[5])&0x80) {
        EffectParam[channel]=(EffectParam[channel]^0xff)+1;
    }
    EffectTempo[channel]=(Instrument[InstrNum[channel]]->Param[7])&0x3f;
    EffectToTempo[channel]=EffectTempo[channel];
#endif
}

/**********************************************************
 funkcje zewnetrzne
 **********************************************************/

//ladowanie songu

zint32 tmc_LoadFile(zchar *FileName)
{
  zint32  tm;
  zint32  MinimalInstAddr;                                  //minimalny adres instrumentu
  zbyte   LoAddr,HiAddr;                                    //zmienne pomocnicze zawierajace niski
                                                            //  i wysoki bajt adresu

  MinimalInstAddr=0xffff;                                   //minimalny adres instrumentu
  if (TMCData!=NULL) {return -1;}                           //  wracamy z bledem gdy nie
                                                            //  wyladowano poprzedniego songu
  if ((TMCFile=fopen(FileName,"rb"))==NULL) {return -1;}    //probujemy otwiorzyc plik z
                                                            //  danymi songu
  if ((fgetc(TMCFile)!=0x00ff)||(fgetc(TMCFile)!=0x00ff))   //sprawdzamy 2 pierwsze bajty
  {                                                         //  musza one byc rowne 0xffff
    fclose(TMCFile);
    return -1;
  }
  TMCBegin=fgetc(TMCFile)+(fgetc(TMCFile)<<8);              //czytamy 6 nastepnych bajtow i
  TMCEnd=fgetc(TMCFile)+(fgetc(TMCFile)<<8);                //  ustawiamy kilka zmiennych
  TMCSize=TMCEnd-TMCBegin+1;                                //  potrzebnych przy obliczaniu
                                                            //  adresow instrumentow i patternow
  TMCData=malloc(TMCSize);                                  //alokujemy pamiec dla danych songu
  for(zint32 i=0;i<TMCSize;i++) {                           //wypelniamy ja danymi z pliku
    *(((zbyte *)TMCData)+i)=(zbyte)(fgetc(TMCFile));
  }
  fclose(TMCFile);                                          //juz wiecej nie bedziemy potrzebowali
                                                            //  pliku wiec go zamykamy
  SongInfo=(struct SongInfoT *)TMCData;                     //ustawiamy kilka zmiennych
  SongData=(struct SongPosT (*)[0x80][0x08])                //  informacjami z pliku
    (((zbyte *)TMCData)+0x01a0);
  for (zint32 i=0;i<0x40;i++) {                             //ustawiamy adresy instrumentow i przy
    LoAddr=SongInfo->LoInstrAddr[i];                        //okazji badamy ktory z nich lezy w
    HiAddr=SongInfo->HiInstrAddr[i];                        //pamieci najnizej
    if ((LoAddr|HiAddr)==0x00) {
      Instrument[i]=&EmptyInstrument;
    } else {
      tm=((((int)HiAddr)<<8)+(int)LoAddr)-TMCBegin;
      Instrument[i]=(InstrT *)(((zbyte *)TMCData)+tm);
      if (tm<MinimalInstAddr) {MinimalInstAddr=tm;}
    }
  }
  for(int i=0;i<0x80;i++) {                                 //ustawiamy adresy patternow
    LoAddr=SongInfo->LoPattAddr[i];
    HiAddr=SongInfo->HiPattAddr[i];
    if ((LoAddr|HiAddr)==0x00) {Pattern[i]=(zbyte (*)[0x0101])
      &EmptyPattern;
    } else {
      Pattern[i]=
        (zbyte (*)[0x0101])(((zbyte *)TMCData)
        -TMCBegin+((((int)HiAddr)<<8)+(int)LoAddr));
    }
  }
  SongLength=(MinimalInstAddr-0x01a0)>>4;                   //ustawiamy ilosc wpisow w songu
  RealSongLength=SongLength;                                //a tutaj szukamy ilosci wpisow
  for(int i=RealSongLength-1;i>=0;i--) {                    //do pierwszego napotkanego skoku
    for(int j=0;j<8;j++) {
      if ((int)((*SongData)[i][j].p)>=0x80) {
        RealSongLength=i;
      }
    }
  }
  for(int i=0x00;i<0x1e;i++) {                              //ustawiamy nazwe songu
    SongName[i]=SongInfo->SongName[i];
  }
  SongName[0x1e]='\0';
  Tempo=SongInfo->FPJ*((SongInfo->Tempo)+1);                //obliczamy i ustawiamy Tempo
  tmc_EvaluateSongForInformation();
  return 0;
}

//uwalnia pamiec i deinicjalizuje song

zvoid tmc_Free(zvoid)
{
    if (TMCData!=NULL) {                                    //uwalniamy pamiec jezeli song jest
        free(TMCData);                                      //zaladowany
        TMCData=NULL;
    }
    SongName[0]='n';                                        //i zerujemy nazwe songu
    SongName[1]='o';
    SongName[2]='n';
    SongName[3]='e';
    SongName[4]='\0';
}

//zerowanie playera

zvoid tmc_Reset(zvoid)
{
    for(int i=0;i<8;i++) {
        InstrCounter[i]=0;
        InstrNum[i]=0xff;
        InstrTune[i]=0;
        AUDFV[i]=0x00;
        AUDCV[i]=0x00;
        PosInPattern[i]=0x00;
        EmptyPos[i]=0x00;
        PriToVolumeDec[i]=0;
        SecToVolumeDec[i]=0;
        PriVolume[i]=0;
        SecVolume[i]=0;
        ChannelPlays[i]=0;
        SubVolume[i]=0x00;
        ToTranspose[i]=0;
        TransposeNum[i]=0;
        FreqTableToUse[i]=0;
        ActualTune[i]=0;
        ChannelFreq[i]=0;
        ChannelTranspose[i]=0;
#ifdef COMPILE_HIGH_PASS_FILTERS
        Audctl[i]=0;
#endif

#ifdef COMPILE_EFFECTS
        DelayToEffect[i]=0;
        EffectHelper1[i]=0;
        EffectHelper2[i]=0;
        EffectParam[i]=0;
        EffectTempo[i]=0;
        EffectToTempo[i]=0;
#endif
    }
#ifdef COMPILE_HIGH_PASS_FILTERS
    AUDCTLV[0]=0;
    AUDCTLV[1]=0;
#endif
    ToTempo=0;
    Tempo=SongInfo->FPJ*((SongInfo->Tempo)+1);
    PatternPos=0x00;
    SongPos=0x00;
}

//zwraca ilosc wywolan na "ramke" playera

zint32 tmc_GetFPJ(zvoid)
{
    return SongInfo->FPJ;
}

//zwraca czas trwania songu w milisekundach

zint32 tmc_GetRealSongLengthInMs(zvoid)
{
  return RealSongLengthInMs;
}

//zwraca dlugosc songu w patternach (wpisach)

zint32 tmc_GetRealSongLength(zvoid)
{
    return RealSongLength;
}

//zwraca wskaznik na lanchuch znakowy w ktorym jest
//  nazwa songu

zchar *tmc_GetSongName(zvoid)
{
    return (char *)SongName;
}

//wypelnia StringToFill nazwa songu z pliku

zint32 tmc_GetFileInfoName(zchar *FileName,zchar *StringToFill)
{
    FILE *TempFile=NULL;

    if ((TempFile=fopen(FileName,"rb"))==NULL) {return -1;}
    if ((fgetc(TempFile)!=0x00ff)||(fgetc(TempFile)!=0x00ff)) {
        return -1;
    }
    fgetc(TempFile);
    fgetc(TempFile);
    fgetc(TempFile);
    fgetc(TempFile);
    for(int i=0;i<0x1d;i++) {
        StringToFill[i]=(char)(fgetc(TempFile));
    }
    StringToFill[0x1d]='\0';
    fclose(TempFile);
    return 0;
}

//jedno wywolanie playera

zint32 tmc_PlayerCall(zvoid)
{
    zbyte z,argument;
    zbyte t,tt,tp,ttt;
    int LeavePatternProc;
    int ReturnValue;
    zbyte effect,param,helper2;
    zbyte TempPriVolume[8];
    zbyte TempSecVolume[8];
    zbyte TempVolume[8];
    zbyte RealChannel[8];

    //przetwarzanie danych songu i patternow
    ReturnValue=0;
    if (ToTempo==0) {
        ttt=0;
        for(int j=0;j<8;j++) {
            ttt=ttt|((*SongData)[SongPos][j].p);
        }
        if (ttt&0x80) {
            return -2;
        }
        if (SongPos>=SongLength) {
            return -1;
        }
        for(int i=7;i>=0;i--) {
            LeavePatternProc=0;
            if (EmptyPos[i]>0x00) {
                EmptyPos[i]--;
                LeavePatternProc=1;
            }
            while (!LeavePatternProc) {
                t=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                if (t==0x00) {
                    PosInPattern[i]++;
                    tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                    SubVolume[i]=tt;
                    PosInPattern[i]++;
                    LeavePatternProc=1;
                }
                if ((t>=0x01)&&(t<=0x3f)) {
                    tmc_PlayInstrument(i,t,(*SongData)[SongPos][7-i].t);
                    PosInPattern[i]++;
                    tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                    SubVolume[i]=tt;
                    PosInPattern[i]++;
                    LeavePatternProc=1;
                }
                if (t==0x40) {
                    PosInPattern[i]++;
                    tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                    if (tt&0x7f) {
                        Tempo=((tt&0x7f)+1)*SongInfo->FPJ;
                        ToTempo=0;
                    } else {
                        PatternPos=0x40;
                    }
                    if (tt&0x80) {
                        PosInPattern[i]++;
                        tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                        SubVolume[i]=tt;
                    }
                    PosInPattern[i]++;
                    LeavePatternProc=1;
                }
                if ((t>=0x41)&&(t<=0x7f)) {
                    tmc_PlayInstrument(i,t,(*SongData)[SongPos][7-i].t);
                    PosInPattern[i]++;
                    tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                    if (tt&0x7f) {
                        Tempo=((tt&0x7f)+1)*SongInfo->FPJ;
                        ToTempo=0;
                    } else {
                        PatternPos=0x40;
                    }
                    PosInPattern[i]++;
                    tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                    SubVolume[i]=tt;
                    PosInPattern[i]++;
                    LeavePatternProc=1;
                }
                if ((t>=0x80)&&(t<=0xbf)) {
                    tmc_ChangeInstrument(i,t&0x3f);
                    PosInPattern[i]++;
                }
                if ((t>=0xc0)&&(t<=0xff)) {
                    EmptyPos[i]=t&0x3f;
                    PosInPattern[i]++;
                    LeavePatternProc=1;
                }
            }
        }
    }
    //przetwarzanie danych instrumentow
    for(int i=7;i>=0;i--) {
        if (ChannelPlays[i]) {
            tp=0;
            if (InstrCounter[i]>20) {
#ifdef COMPILE_HIGH_PASS_FILTERS
                Audctl[i]=Instrument[InstrNum[i]]->Param[1];
#endif
                InstrCounter[i]=20;
                tp=1;
                //opadanie glownego kanalu
                if (Instrument[InstrNum[i]]->Param[3]!=0) {
                    if (++PriToVolumeDec[i]>=Instrument[InstrNum[i]]->Param[3]) {
                        PriToVolumeDec[i]=0;
                        if (PriVolume[i]>0) {PriVolume[i]--;}
                    }
                }
                //opadanie dodatkowego kanalu
                if (Instrument[InstrNum[i]]->Param[4]!=0) {
                    if (++SecToVolumeDec[i]>=Instrument[InstrNum[i]]->Param[4]) {
                        SecToVolumeDec[i]=0;
                        if (SecVolume[i]>0) {SecVolume[i]--;}
                    }
                }
            } else {
                PriVolume[i]=(Instrument[InstrNum[i]]->value[InstrCounter[i]].primary)&0x0f;
                SecVolume[i]=(Instrument[InstrNum[i]]->value[InstrCounter[i]].secondary)&0x0f;
                ChannelDist[i]=(Instrument[InstrNum[i]]->value[InstrCounter[i]].primary)&0xf0;
                z=(Instrument[InstrNum[i]]->value[InstrCounter[i]].secondary)&0xf0;
                argument=Instrument[InstrNum[i]]->value[InstrCounter[i]].argument;
#ifdef COMPILE_EFFECTS
                EffectHelper1[i]=0;
#endif
#ifdef COMPILE_HIGH_PASS_FILTERS
                if (z) {
                    if (z&0x80) {
                        Audctl[i]=Instrument[InstrNum[i]]->Param[2];
                    } else {
                        Audctl[i]=0;
                    }
                } else {
                    Audctl[i]=Instrument[InstrNum[i]]->Param[1];
                }
#endif
                z=z&0x70;
                if (z==0x00) {
                    tp=1;
                }
                if (z==0x10) {
                    ChannelFreq[i]=argument;
                }
                if (z==0x20) {
                    ChannelFreq[i]+=argument;
                }
                if (z==0x30) {
                    ActualTune[i]=InstrTune[i]
                        +FreqTableToUse[i]
                        +ChannelTranspose[i];
                    ChannelFreq[i]=tmc_TuneToFreq(
                        ActualTune[i])+argument;
                }
                if (z==0x40) {
                    ChannelFreq[i]=argument&(rand()%256);
                }
                if (z==0x50) {
                    ActualTune[i]=argument;
                    ChannelFreq[i]=tmc_TuneToFreq(
                        ActualTune[i]);
                }
                if (z==0x60) {
                    ChannelFreq[i]+=tmc_TuneToFreq(argument);
                }
                if (z==0x70) {
                    ActualTune[i]=InstrTune[i]
                        +FreqTableToUse[i]+argument
                        +ChannelTranspose[i];
                    ChannelFreq[i]=tmc_TuneToFreq(
                        ActualTune[i]);
                }
            }
            if (tp) {
                ActualTune[i]=InstrTune[i]+FreqTableToUse[i]
                    +ChannelTranspose[i]
                    +Instrument[InstrNum[i]]->Transpos[TransposeNum[i]];
                ChannelFreq[i]=tmc_TuneToFreq(ActualTune[i]);
                if (++ToTranspose[i]>(Instrument[InstrNum[i]]->Param[8]&0x0f)) {
                    ToTranspose[i]=0;
                    if (Instrument[InstrNum[i]]->Param[8]&0x80) {
                        if (--TransposeNum[i]<0) {
                            TransposeNum[i]=((Instrument[InstrNum[i]]->Param[8]&0x70)>>4);
                        }
                    } else {
                        if (++TransposeNum[i]>((Instrument[InstrNum[i]]->Param[8]&0x70)>>4)) {
                            TransposeNum[i]=0;
                        }
                    }
                }
#ifdef COMPILE_EFFECTS
                if (DelayToEffect[i]) {
                    --DelayToEffect[i];
                } else {
                    effect=(Instrument[InstrNum[i]]->Param[5])&0x70;
                    helper2=EffectHelper2[i];
                    param=EffectParam[i];
                    if (effect==0x10) {
                        if (helper2&0x01) {
                            if (helper2&0x02) {
                                EffectHelper1[i]=-param;
                                ChannelFreq[i]-=param;
                            } else {
                                EffectHelper1[i]=param;
                                ChannelFreq[i]+=param;
                            }
                        } else {
                            EffectHelper1[i]=0;
                        }
                        helper2++;
                    }
                    if (effect==0x20) {
                        if (param&0x80) {helper2--;} else {helper2++;}
                        if (param==helper2) {param^=0xff;}
                        EffectHelper1[i]=helper2;
                        ChannelFreq[i]+=helper2;
                    }
                    if (effect==0x30) {
                        if (param&0x80) {helper2--;} else {helper2++;}
                        if (param==helper2) {param^=0xff;}
                        ActualTune[i]+=helper2;
                        ChannelFreq[i]=tmc_TuneToFreq(ActualTune[i]);
                    }
                    if (effect==0x40) {
                        EffectHelper1[i]=helper2;
                        ChannelFreq[i]+=helper2;
                        helper2+=param;
                    }
                    if (effect==0x50) {
                        ActualTune[i]-=helper2;
                        ChannelFreq[i]=tmc_TuneToFreq(ActualTune[i]);
                        helper2+=param;
                    }
                    if (effect==0x60) {
                        ChannelFreq[i]-=param;
                        EffectHelper1[i]=-param;
                    }
                    if (effect==0x70) {
                        //uwaga, nie przetestowalem tego efektu
                        if (!EffectToTempo[i]) {
                            if (param&0x80) {
                                if ((SecVolume[i]!=0) && (PriVolume[i]!=0x0f)) {
                                    PriVolume[i]++;
                                }
                            } else {
                                if ((PriVolume[i]!=0) && (SecVolume[i]!=0x0f)) {
                                    SecVolume[i]++;
                                }
                            }
                        }
                    }
                    if ((--EffectToTempo[i])&0x80) {
                        EffectHelper2[i]=helper2;
                        EffectParam[i]=param;
                        EffectToTempo[i]=EffectTempo[i];
                    }
                }
#endif
            }
#ifdef COMPILE_HIGH_PASS_FILTERS
            if ((i==0) && (Audctl[0]&0x04)) {
                ActualTune[2]=ActualTune[0]+Instrument[InstrNum[i]]->Param[0];
                ChannelFreq[2]=tmc_TuneToFreq(ActualTune[2])
                    +EffectHelper1[0]+1;
            }
            if ((i==1) && (Audctl[1]&0x02)) {
                ActualTune[3]=ActualTune[1]+Instrument[InstrNum[i]]->Param[0];
                ChannelFreq[3]=tmc_TuneToFreq(ActualTune[3])
                    +EffectHelper1[1]+1;
            }
            if ((i==4) && (Audctl[4]&0x04)) {
                ActualTune[6]=ActualTune[4]+Instrument[InstrNum[i]]->Param[0];
                ChannelFreq[6]=tmc_TuneToFreq(ActualTune[6])
                    +EffectHelper1[4]+1;
            }
            if ((i==5) && (Audctl[5]&0x02)) {
                ActualTune[7]=ActualTune[5]+Instrument[InstrNum[i]]->Param[0];
                ChannelFreq[7]=tmc_TuneToFreq(ActualTune[7])
                    +EffectHelper1[5]+1;
            }
#endif
            InstrCounter[i]++;
        } else {
            ActualTune[i]=0;
            ChannelFreq[i]=0;
            ChannelDist[i]=0;
            PriVolume[i]=0;
            SecVolume[i]=0;
#ifdef COMPILE_HIGH_PASS_FILTERS
            Audctl[i]=0x00;
#endif
        }
    }
    for(int i=7;i>=0;i--) {
        t=PriVolume[i]-((SubVolume[i]&0xf0)>>4);
        if (t&0xf0) {t=0x00;}
        if (!(ActualTune[i]&0x3f)) {t=0;}
        TempPriVolume[i]=t;
        t=SecVolume[i]-(SubVolume[i]&0x0f);
        if (t&0xf0) {t=0x00;}
        if (!(ActualTune[i]&0x3f)) {t=0;}
        TempSecVolume[i]=t;
    }
    for(int i=7;i>=0;i--) {
        if ((!TempPriVolume[i]) && TempSecVolume[InversedChannel[i]]) {
            RealChannel[i]=InversedChannel[i];
            TempVolume[i]=TempSecVolume[RealChannel[i]];
        } else {
            RealChannel[i]=i;
            TempVolume[i]=TempPriVolume[i];
        }
        AUDFV[i]=ChannelFreq[RealChannel[i]];
#ifdef COMPILE_HIGH_PASS_FILTERS
        if ((i==0) && (Audctl[RealChannel[0]]&0x04)) {
            AUDFV[2]=ChannelFreq[RealChannel[2]];
        }
        if ((i==1) && (Audctl[RealChannel[1]]&0x02)) {
            AUDFV[3]=ChannelFreq[RealChannel[3]];
        }
        if ((i==4) && (Audctl[RealChannel[4]]&0x04)) {
            AUDFV[6]=ChannelFreq[RealChannel[6]];
        }
        if ((i==5) && (Audctl[RealChannel[5]]&0x02)) {
            AUDFV[7]=ChannelFreq[RealChannel[7]];
        }
#endif
        AUDCV[i]=(ChannelDist[RealChannel[i]]&0xf0)+(TempVolume[i]&0x0f);
    }
    if ((Audctl[RealChannel[1]]&0x10)) {
        AUDFV[1]=Hi16Table[ActualTune[RealChannel[1]]&0x3f];
        AUDFV[0]=Lo16Table[ActualTune[RealChannel[1]]&0x3f];
    }
    if ((Audctl[RealChannel[3]]&0x08)) {
        AUDFV[3]=Hi16Table[ActualTune[RealChannel[3]]&0x3f];
        AUDFV[2]=Lo16Table[ActualTune[RealChannel[3]]&0x3f];
    }
    if ((Audctl[RealChannel[5]]&0x10)) {
        AUDFV[5]=Hi16Table[ActualTune[RealChannel[5]]&0x3f];
        AUDFV[4]=Lo16Table[ActualTune[RealChannel[5]]&0x3f];
    }
    if ((Audctl[RealChannel[7]]&0x08)) {
        AUDFV[7]=Hi16Table[ActualTune[RealChannel[7]]&0x3f];
        AUDFV[6]=Lo16Table[ActualTune[RealChannel[7]]&0x3f];
    }
    AUDCTLV[0]=0;
    AUDCTLV[1]=0;
#ifdef COMPILE_HIGH_PASS_FILTERS
    for(int i=0;i<4;i++) {
        AUDCTLV[0]=AUDCTLV[0]|Audctl[RealChannel[i]];
        AUDCTLV[1]=AUDCTLV[1]|Audctl[RealChannel[i+4]];
    }
#endif
    if (ToTempo==0) {
        PatternPos++;
        if (PatternPos>=0x40) {
            PatternPos=0x00;
            for(int i=0;i<8;i++) {
                EmptyPos[i]=0x00;
                PosInPattern[i]=0;
            }
            SongPos++;
        }
    }
    ToTempo++;
    if (ToTempo>=Tempo) {ToTempo=0;}
    return ReturnValue;
}

//skok przy koncu poisenki lub przy napotkaniu rozkazu skoku

zvoid tmc_MakeJump(zvoid)
{
    if ((int)((*SongData)[SongPos][7].p)==0xff) {
        SongPos=(*SongData)[SongPos][7].t;
    } else {
        SongPos=0;
    }
    ToTempo=0;
    PatternPos=0x00;
    for(int i=0;i<8;i++) {
        EmptyPos[i]=0x00;
        PosInPattern[i]=0;
    }
}

//skacze do podanej pozycji w songu

zvoid tmc_MakeJumpTo(zbyte NewSongPos,zbyte NewPatternPos)
{
/*    zbyte t,tt;
    int LeavePatternProc;

    if (NewPatternPos>=0x40) {NewPatternPos=0x3f;}
    if (NewSongPos>RealSongLength) {
        NewSongPos=RealSongLength;
        NewPatternPos=0x00;
    }
    SongPos=NewSongPos;
    PatternPos=NewPatternPos;
    ToTempo=0;
    for(int i=0;i<8;i++) {
        EmptyPos[i]=0;
        PosInPattern[i]=0;
    }
    if (NewPatternPos!=0x00) {
        for(int j=0;j<PatternPos;j++) {
            for(int i=0;i<8;i++) {
                LeavePatternProc=0;
                if (EmptyPos[i]>0x00) {
                    EmptyPos[i]--;
                    LeavePatternProc=1;
                }
                while (!LeavePatternProc) {
                    t=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                    if (t==0x00) {
                        PosInPattern[i]+=2;
                        LeavePatternProc=1;
                    }
                    if ((t>=0x01)&&(t<=0x3f)) {
                        PosInPattern[i]+=2;
                        LeavePatternProc=1;
                    }
                    if (t==0x40) {
                        PosInPattern[i]++;
                        tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                        if (tt&0x7f) {
                            Tempo=((tt&0x7f)+1)*SongInfo->FPJ;
                            ToTempo=0;
                        } else {
                            PatternPos=0x40;
                        }
                        if (tt&0x80) {
                            PosInPattern[i]++;
                            tt=(*(Pattern[(*SongData)[SongPos][7-i].p]))[PosInPattern[i]];
                            SubVolume[i]=tt;
                        }
                        PosInPattern[i]++;
                        LeavePatternProc=1;
                    }
                    if ((t>=0x41)&&(t<=0x7f)) {
                        PosInPattern[i]+=3;
                        LeavePatternProc=1;
                    }
                    if ((t>=0x80)&&(t<=0xbf)) {
                        PosInPattern[i]+=1;
                    }
                    if ((t>=0xc0)&&(t<=0xff)) {
                        EmptyPos[i]=t&0x3f;
                        PosInPattern[i]+=1;
                        LeavePatternProc=1;
                    }
                }
            }
        }
    }*/
}

//skacze do podanej pozycji w songu w milisekundach

zvoid tmc_MakeJumpToMs(zint32 ToMs)
{
    int t,tt;

    tmc_Reset();
    ToMs*=SongInfo->FPJ;
    t=20;
    tt=0;
    while ((tt+=t)<ToMs) {tmc_PlayerCall();}
}

//zwraca AUDF ktore powinno byc przekazane do pokeya

zbyte tmc_GetAUDF(zbyte channel)
{
    return AUDFV[channel];
}

//zwraca AUDC ktore powinno byc przekazane do pokeya

zbyte tmc_GetAUDC(zbyte channel)
{
    return AUDCV[channel];
}

//zwraca AUDCTL ktore powinno byc przekazane do pokeya

zbyte tmc_GetAUDCTL(zbyte pokey)
{
  return AUDCTLV[pokey];
}

#endif

