#ifndef TMCPLAY_H
#define TMCPLAY_H

#include "ztypes.h"                                         //common atari data types
#include "compile.h"                                        //compile options

#ifdef TMC_COMPILE

#define TMC_V1_PLAYER_VER   0.90
//remove this to compile without effects
#define COMPILE_EFFECTS
//remove this to compile without high-pass filtering
#define COMPILE_HIGH_PASS_FILTERS

/**********************************************************
  funkcja laduje plik TMC do pamieci i zwraca zero jezeli
  odczyt sie powiodl, jezeli nie to zwraca wartosc ujemna
 **********************************************************/
zint32 tmc_LoadFile(zchar * FileName);

/**********************************************************
  uwalnia pamiec zajeta przez song
 **********************************************************/
zvoid tmc_Free(zvoid);

/**********************************************************
  zeruje liczniki wykonania songu i wycisza kanaly
 **********************************************************/
zvoid tmc_Reset(zvoid);

/**********************************************************
  zwraca ilosc wywolan na "ramke" playera
 **********************************************************/
zint32 tmc_GetFPJ(zvoid);

/**********************************************************
  zwraca czas trwania songu w milisekundach
 **********************************************************/
zint32 tmc_GetRealSongLengthInMs(zvoid);

/**********************************************************
  zwraca dlugosc songu w patternach (wpisach)
 **********************************************************/
zint32 tmc_GetRealSongLength(zvoid);

/**********************************************************
  zwraca wskaznik na lanchuch znakowy w ktorym jest
  nazwa songu
 **********************************************************/
zchar *tmc_GetSongName(zvoid);

/**********************************************************
  wypelnia StringToFill nazwa songu ktory znajdzie w
  podanym pliku, nazwa ma zawsze 30 znakow, ostatni to \0
  funkcja zwraca 0 jezeli sie wszystko powiodlo lub
  -1 jezeli nie moze znalezc pliku lub go nie rozpoznaje
 **********************************************************/
zint32 tmc_GetFileInfoName(zchar *FileName,zchar *StringToFill);

/**********************************************************
  jedno wywolanie playera, zmieniana jest tu faza
  odgrywanych instrumentow i jezeli potrzeba takze posycji
  w patternie i songu
  funkcja zwraca:
  0  - jezeli wszystko oki
  -1 - jezeli natrafiono na koniec songu
  -2 - jezeli natrafiono na rozkaz skoku w songu
  przy wartosciach -1 i -2 player nie bedzie puszczal
  muzyczki dalej, za kazdym wywolaniem bedzie zwracal ten
  sam blad do czasu wywolania funkcji ResetTMC lub
  MakeJump
 **********************************************************/
zint32 tmc_PlayerCall(zvoid);

/**********************************************************
  jezeli NextTMCFaze zwrocil -1 lub -2 to mozna puscic
  song dalej za pomoca wlasnie tej funkcji
  jezeli player napotkal na koniec songu to ta funkcja
  spowoduje skok do poczatku songu
  jezeli player napotkal rozkaz skoku to ta funkcja go
  wykona
  uwaga!!! funkcja ta nie resetuje instrumentow, powoduje
  tylko skok kursora odtwarzania
 **********************************************************/
zvoid tmc_MakeJump(zvoid);

/**********************************************************
  skacze do podanej pozycji w songu
  uwaga!!! funkcja ta nie resetuje instrumentow, powoduje
  tylko skok kursora odtwarzania
 **********************************************************/
zvoid tmc_MakeJumpTo(zbyte NewSongPos,zbyte NewPatternPos);

/**********************************************************
  skacze do podanej pozycji w songu w milisekundach
  uwaga!!! funkcja ta nie resetuje instrumentow, powoduje
  tylko skok kursora odtwarzania
 **********************************************************/
zvoid tmc_MakeJumpToMs(zint32 ToMs);

/**********************************************************
  zwraca AUDF ktore powinno byc przekazane do pokeya
 **********************************************************/
zbyte tmc_GetAUDF(zbyte channel);

/**********************************************************
  zwraca AUDC ktore powinno byc przekazane do pokeya
 **********************************************************/
zbyte tmc_GetAUDC(zbyte channel);

/**********************************************************
 zwraca AUDCTL ktore powinno byc przekazane do pokeya
 **********************************************************/
zbyte tmc_GetAUDCTL(zbyte pokey);

#endif                                                      //of #ifdef TMC_COMPILE
#endif

